from ... import platforms
from .. import state
from . import cpu


class MIPS(cpu.CPU):
    """Auto-generated CPU state for mips:mips32:big.

    Generated from Pcode language MIPS:BE:32:default, and Unicorn package
    unicorn.mips_const.
    """

    def __init__(self):
        super().__init__()
        # NOTE: MIPS registers have both a name and a number.

        # *** General-Purpose Registers ***
        # Assembler-Temporary Register
        self.at = state.Register("at", size=4)
        self.add(self.at)
        self._1 = state.RegisterAlias("1", self.at, size=4, offset=0)
        self.add(self._1)
        # Return Value Registers
        self.v0 = state.Register("v0", size=4)
        self.add(self.v0)
        self._2 = state.RegisterAlias("2", self.v0, size=4, offset=0)
        self.add(self._2)
        self.v1 = state.Register("v1", size=4)
        self.add(self.v1)
        self._3 = state.RegisterAlias("3", self.v1, size=4, offset=0)
        self.add(self._3)
        # Argument Registers
        self.a0 = state.Register("a0", size=4)
        self.add(self.a0)
        self._4 = state.RegisterAlias("4", self.a0, size=4, offset=0)
        self.add(self._4)
        self.a1 = state.Register("a1", size=4)
        self.add(self.a1)
        self._5 = state.RegisterAlias("5", self.a1, size=4, offset=0)
        self.add(self._5)
        self.a2 = state.Register("a2", size=4)
        self.add(self.a2)
        self._6 = state.RegisterAlias("6", self.a2, size=4, offset=0)
        self.add(self._6)
        self.a3 = state.Register("a3", size=4)
        self.add(self.a3)
        self._7 = state.RegisterAlias("7", self.a3, size=4, offset=0)
        self.add(self._7)
        # Temporary Registers
        self.t0 = state.Register("t0", size=4)
        self.add(self.t0)
        self._8 = state.RegisterAlias("8", self.t0, size=4, offset=0)
        self.add(self._8)
        self.t1 = state.Register("t1", size=4)
        self.add(self.t1)
        self._9 = state.RegisterAlias("9", self.t1, size=4, offset=0)
        self.add(self._9)
        self.t2 = state.Register("t2", size=4)
        self.add(self.t2)
        self._10 = state.RegisterAlias("10", self.t2, size=4, offset=0)
        self.add(self._10)
        self.t3 = state.Register("t3", size=4)
        self.add(self.t3)
        self._11 = state.RegisterAlias("11", self.t3, size=4, offset=0)
        self.add(self._11)
        self.t4 = state.Register("t4", size=4)
        self.add(self.t4)
        self._12 = state.RegisterAlias("12", self.t4, size=4, offset=0)
        self.add(self._12)
        self.t5 = state.Register("t5", size=4)
        self.add(self.t5)
        self._13 = state.RegisterAlias("13", self.t5, size=4, offset=0)
        self.add(self._13)
        self.t6 = state.Register("t6", size=4)
        self.add(self.t6)
        self._14 = state.RegisterAlias("14", self.t6, size=4, offset=0)
        self.add(self._14)
        self.t7 = state.Register("t7", size=4)
        self.add(self.t7)
        self._15 = state.RegisterAlias("15", self.t7, size=4, offset=0)
        self.add(self._15)
        # NOTE: These numbers aren't out of order.
        # t8 and t9 are later in the register file than t0 - t7.
        self.t8 = state.Register("t8", size=4)
        self.add(self.t8)
        self._24 = state.RegisterAlias("24", self.t8, size=4, offset=0)
        self.add(self._24)
        self.t9 = state.Register("t9", size=4)
        self.add(self.t9)
        self._25 = state.RegisterAlias("25", self.t9, size=4, offset=0)
        self.add(self._25)
        # Saved Registers
        self.s0 = state.Register("s0", size=4)
        self.add(self.s0)
        self._16 = state.RegisterAlias("16", self.s0, size=4, offset=0)
        self.add(self._16)
        self.s1 = state.Register("s1", size=4)
        self.add(self.s1)
        self._17 = state.RegisterAlias("17", self.s1, size=4, offset=0)
        self.add(self._17)
        self.s2 = state.Register("s2", size=4)
        self.add(self.s2)
        self._18 = state.RegisterAlias("18", self.s2, size=4, offset=0)
        self.add(self._18)
        self.s3 = state.Register("s3", size=4)
        self.add(self.s3)
        self._19 = state.RegisterAlias("19", self.s3, size=4, offset=0)
        self.add(self._19)
        self.s4 = state.Register("s4", size=4)
        self.add(self.s4)
        self._20 = state.RegisterAlias("20", self.s4, size=4, offset=0)
        self.add(self._20)
        self.s5 = state.Register("s5", size=4)
        self.add(self.s5)
        self._21 = state.RegisterAlias("21", self.s5, size=4, offset=0)
        self.add(self._21)
        self.s6 = state.Register("s6", size=4)
        self.add(self.s6)
        self._22 = state.RegisterAlias("22", self.s6, size=4, offset=0)
        self.add(self._22)
        self.s7 = state.Register("s7", size=4)
        self.add(self.s7)
        self._23 = state.RegisterAlias("23", self.s7, size=4, offset=0)
        self.add(self._23)
        # NOTE: Register #30 was originally the Frame Pointer.
        # It's been re-aliased as s8, since many ABIs don't use the frame pointer.
        # Unicorn and Sleigh prefer to use the alias s8,
        # so it should be the base register.
        self.s8 = state.Register("s8", size=4)
        self.add(self.s8)
        self.fp = state.RegisterAlias("fp", self.s8, size=4, offset=0)
        self.add(self.fp)
        self._30 = state.RegisterAlias("30", self.s8, size=4, offset=0)
        self.add(self._30)
        # Kernel-reserved Registers
        self.k0 = state.Register("k0", size=4)
        self.add(self.k0)
        self._26 = state.RegisterAlias("26", self.k0, size=4, offset=0)
        self.add(self._26)
        self.k1 = state.Register("k1", size=4)
        self.add(self.k1)
        self._27 = state.RegisterAlias("27", self.k1, size=4, offset=0)
        self.add(self._27)
        # *** Pointer Registers ***
        # Zero register
        self.zero = state.FixedRegister("zero", size=4, value=0)
        self.add(self.zero)
        self._0 = state.RegisterAlias("0", self.zero, size=4, offset=0)
        self.add(self._0)
        # Global Offset Pointer
        self.gp = state.Register("gp", size=4)
        self.add(self.gp)
        self._28 = state.RegisterAlias("28", self.gp, size=4, offset=0)
        self.add(self._28)
        # Stack Pointer
        self.sp = state.Register("sp", size=4)
        self.add(self.sp)
        self._29 = state.RegisterAlias("29", self.sp, size=4, offset=0)
        self.add(self._29)
        # Return Address
        self.ra = state.Register("ra", size=4)
        self.add(self.ra)
        self._31 = state.RegisterAlias("31", self.ra, size=4, offset=0)
        self.add(self._31)
        # Program Counter
        self.pc = state.Register("pc", size=4)
        self.add(self.pc)
        # Floating Point Registers
        self.f0 = state.Register("f0", size=8)
        self.add(self.f0)
        self.f1 = state.Register("f1", size=8)
        self.add(self.f1)
        self.f2 = state.Register("f2", size=8)
        self.add(self.f2)
        self.f3 = state.Register("f3", size=8)
        self.add(self.f3)
        self.f4 = state.Register("f4", size=8)
        self.add(self.f4)
        self.f5 = state.Register("f5", size=8)
        self.add(self.f5)
        self.f6 = state.Register("f6", size=8)
        self.add(self.f6)
        self.f7 = state.Register("f7", size=8)
        self.add(self.f7)
        self.f8 = state.Register("f8", size=8)
        self.add(self.f8)
        self.f9 = state.Register("f9", size=8)
        self.add(self.f9)
        self.f10 = state.Register("f10", size=8)
        self.add(self.f10)
        self.f11 = state.Register("f11", size=8)
        self.add(self.f11)
        self.f12 = state.Register("f12", size=8)
        self.add(self.f12)
        self.f13 = state.Register("f13", size=8)
        self.add(self.f13)
        self.f14 = state.Register("f14", size=8)
        self.add(self.f14)
        self.f15 = state.Register("f15", size=8)
        self.add(self.f15)
        self.f16 = state.Register("f16", size=8)
        self.add(self.f16)
        self.f17 = state.Register("f17", size=8)
        self.add(self.f17)
        self.f18 = state.Register("f18", size=8)
        self.add(self.f18)
        self.f19 = state.Register("f19", size=8)
        self.add(self.f19)
        self.f20 = state.Register("f20", size=8)
        self.add(self.f20)
        self.f21 = state.Register("f21", size=8)
        self.add(self.f21)
        self.f22 = state.Register("f22", size=8)
        self.add(self.f22)
        self.f23 = state.Register("f23", size=8)
        self.add(self.f23)
        self.f24 = state.Register("f24", size=8)
        self.add(self.f24)
        self.f25 = state.Register("f25", size=8)
        self.add(self.f25)
        self.f26 = state.Register("f26", size=8)
        self.add(self.f26)
        self.f27 = state.Register("f27", size=8)
        self.add(self.f27)
        self.f28 = state.Register("f28", size=8)
        self.add(self.f28)
        self.f29 = state.Register("f29", size=8)
        self.add(self.f29)
        self.f30 = state.Register("f30", size=8)
        self.add(self.f30)
        self.f31 = state.Register("f31", size=8)
        self.add(self.f31)
        # *** Floating Point Control Registers ***
        # NOTE: These are taken from Sleigh, and the MIPS docs.
        # Unicorn doesn't use these names, and has a different number of registers.
        self.fir = state.Register("fir", size=4)
        self.add(self.fir)
        self.fcsr = state.Register("fcsr", size=4)
        self.add(self.fcsr)
        self.fexr = state.Register("fexr", size=4)
        self.add(self.fexr)
        self.fenr = state.Register("fenr", size=4)
        self.add(self.fenr)
        self.fccr = state.Register("fccr", size=4)
        self.add(self.fccr)
        # TODO: MIPS has a boatload of extensions with their own registers.
        # There isn't a clean join between Sleigh, Unicorn, and MIPS docs.


class MIPSEL(MIPS):
    """Auto-generated CPU state for mips:mips32:little.

    Generated from Pcode language MIPS:LE:32:default, and Unicorn package
    unicorn.mips_const.
    """

    platform = platforms.Platform(
        platforms.Architecture.MIPS32, platforms.Byteorder.LITTLE
    )

    def __init__(self):
        super().__init__()
        # *** Accumulator Registers ***
        # MIPS uses these to implement 64-bit results
        # from 32-bit multiplication, amongst others.
        self.ac0 = state.Register("ac0", size=8)
        self.add(self.ac0)
        self.lo0 = state.RegisterAlias("lo0", self.ac0, size=4, offset=0)
        self.add(self.lo0)
        self.hi0 = state.RegisterAlias("hi0", self.ac0, size=4, offset=4)
        self.add(self.hi0)
        self.ac1 = state.Register("ac1", size=8)
        self.add(self.ac1)
        self.lo1 = state.RegisterAlias("lo1", self.ac1, size=4, offset=0)
        self.add(self.lo1)
        self.hi1 = state.RegisterAlias("hi1", self.ac1, size=4, offset=4)
        self.add(self.hi1)
        self.ac2 = state.Register("ac2", size=8)
        self.add(self.ac2)
        self.lo2 = state.RegisterAlias("lo2", self.ac2, size=4, offset=0)
        self.add(self.lo2)
        self.hi2 = state.RegisterAlias("hi2", self.ac2, size=4, offset=4)
        self.add(self.hi2)
        self.ac3 = state.Register("ac3", size=8)
        self.add(self.ac3)
        self.lo3 = state.RegisterAlias("lo3", self.ac3, size=4, offset=0)
        self.add(self.lo3)
        self.hi3 = state.RegisterAlias("hi3", self.ac3, size=4, offset=4)
        self.add(self.hi3)


class MIPSBE(MIPS):
    """Auto-generated CPU state for mips:mips32:big.

    Generated from Pcode language MIPS:BE:32:default, and Unicorn package
    unicorn.mips_const.
    """

    platform = platforms.Platform(
        platforms.Architecture.MIPS32, platforms.Byteorder.BIG
    )

    def __init__(self):
        super().__init__()
        # *** Accumulator Registers ***
        # MIPS uses these to implement 64-bit results
        # from 32-bit multiplication, amongst others.
        self.ac0 = state.Register("ac0", size=8)
        self.add(self.ac0)
        self.hi0 = state.RegisterAlias("hi0", self.ac0, size=4, offset=0)
        self.add(self.hi0)
        self.lo0 = state.RegisterAlias("lo0", self.ac0, size=4, offset=4)
        self.add(self.lo0)
        self.ac1 = state.Register("ac1", size=8)
        self.add(self.ac1)
        self.hi1 = state.RegisterAlias("hi1", self.ac1, size=4, offset=0)
        self.add(self.hi1)
        self.lo1 = state.RegisterAlias("lo1", self.ac1, size=4, offset=4)
        self.add(self.lo1)
        self.ac2 = state.Register("ac2", size=8)
        self.add(self.ac2)
        self.hi2 = state.RegisterAlias("hi2", self.ac2, size=4, offset=0)
        self.add(self.hi2)
        self.lo2 = state.RegisterAlias("lo2", self.ac2, size=4, offset=4)
        self.add(self.lo2)
        self.ac3 = state.Register("ac3", size=8)
        self.add(self.ac3)
        self.hi3 = state.RegisterAlias("hi3", self.ac3, size=4, offset=0)
        self.add(self.hi3)
        self.lo3 = state.RegisterAlias("lo3", self.ac3, size=4, offset=4)
        self.add(self.lo3)



import logging

import angr
from angr.storage.memory_mixins.memory_mixin import MemoryMixin

from .... import hinting
from ....emulators.angr import PathTerminationSignal
from ....exceptions import AnalysisSignal
from .base import BaseMemoryMixin
from .utils import print_state
from .visitor import ConditionalVisitor

log = logging.getLogger(__name__)


class DivergentAddressSignal(AnalysisSignal):
    """Fault for communicating divergent address data between plugins.

    Divergent address concretizations are detected
    in the memory plugin, but can only be resolved
    in the exploration strategy.
    """

    def __init__(self, state, addr, results):
        self.state = state
        self.addr = addr
        self.results = results


class DivergenceMemoryMixin(BaseMemoryMixin):
    _visitor = ConditionalVisitor()
    """Mixin for handling memory-side address concretization

    This hooks the actual address concretization operation,
    and looks for attempts to concretize conditional addresses.
    angr's default concretization is extremely imprecise
    when applied to conditional expressions,
    so hooking the operation gives us freedom to handle things differently.

    The current strategy I like is to fork the execution state
    into separate copies, one per possible evaluation of the conditional.
    The actual fork operation cannot be done from within this plugin,
    so it needs to communicate with DivergenceExplorationMixin.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @MemoryMixin.memo
    def copy(self, memo):
        o = super().copy(memo)
        o._setup_tui()
        return o

    def _concretize_addr(self, supercall, addr, strategies=None, condition=None):
        """Helper function for performing conditional concretization.

        This replaces the normal concretization strategy,
        which is to compute all possible suggestions for addr,
        and to assert the OR of them.

        This default strategy falls over badly for addresses computed
        via conditional data flow, since said addresses will be
        essentially a case-statement of several possible base addresses.
        The resulting assertion becomes "one of these variables stores one of these addresses."
        This loses a lot of precision.

        This implementation works as follows:

        - Carve apart the address expression into all possible non-conditional expressions
        - Concretize each possible expression independently.
        - If there is more than one possible sub-expression, raise an exception.
        - The exception acts as a virtual pagefault, which carries data back to the exploration strategy.
        - The exploration strategy creates one state per sub-expression, constrained so that said expression is picked.

        N.B.: This is in a separate helper because the operation is identical
        for concretizing reads and writes.
        """
        exprs = dict()
        guards = dict()
        res = set()

        block = self.state.block()
        (insn,) = list(
            filter(
                lambda x: x.address == self.state._ip.concrete_value,
                block.disassembly.insns,
            )
        )

        try:
            # Split the address into non-conditional sub-expressions.
            for expr, guard in self._visitor.visit(addr):
                try:
                    # Concretize the sub-expression normally
                    for tmp in supercall(
                        expr, strategies=strategies, condition=condition
                    ):
                        # check that the suggestion is satisfiable.
                        # I'd have hoped that angr does this automatically,
                        # but it doesn't seem to.
                        if self.state.satisfiable(extra_constraints=[addr == tmp]):
                            log.debug(
                                f"Concrete result for {expr}, slice of {addr}: {tmp:x} -> {self.state.memory.load(tmp, 8)}"
                            )
                            # TODO: Handle multiple concretizations from non-conditional source.
                            # I've never actually seen this, but you never know.
                            guards[expr] = guard
                            exprs[expr] = tmp
                            res.add(tmp)
                        else:
                            log.debug(
                                f"Ignoring result for {expr}, slice of {addr}: {tmp:x}"
                            )
                            for expr in self.state.solver.constraints:
                                log.debug(f"\t{expr}")
                except angr.errors.SimMemoryAddressError as e:
                    # Something went wrong that shouldn't go wrong.
                    log.error(
                        f"Could not concretize expression {expr} with following constraints:"
                    )
                    for expr in self.state.solver.constraints:
                        log.error(f"\t{expr}")
                    log.error(f"Cause: {e}")
                    raise e
        except AnalysisSignal as s:
            # A lower analysis raised a signal.
            # Raise it upwards.
            raise s
        except Exception as e:
            # Something went wrong that shouldn't go wrong.
            log.error(f"Fatal error concretizing {addr}")
            log.exception(f"Cause: {type(e)} {e}")
            raise e
        log.debug("All recommendations:")
        for expr, val in exprs.items():
            log.debug(f"  {expr} := {hex(val)}")
        if len(exprs) > 1:
            # We've got a conditional dereference.
            hint = hinting.UnderSpecifiedMemoryBranchHint(
                message="Conditional address dereference",
                instruction=self.state._ip.concrete_value,
                address=str(addr),
                options=[(str(k), str(v)) for (k, v) in guards.items()],
            )
            self.hinter.info(hint)
            options = {
                "fork": self.divergence_fork,
                "choose": self.divergence_choose,
                "ignore": self.divergence_ignore,
                "details": self.divergence_details,
                "stop": self.divergence_stop,
                "quit": self.divergence_quit,
            }
            option = "fork"
            options[option](this=self, addr=addr, exprs=exprs)

        return list(res)

    def concretize_read_addr(self, addr, strategies=None, condition=None):
        log.debug(f"Concretizing read addr {addr}")
        return self._concretize_addr(
            super().concretize_read_addr,
            addr,
            strategies=strategies,
            condition=condition,
        )

    def concretize_write_addr(self, addr, strategies=None, condition=None):
        log.debug(f"Concretizing write addr {addr}")
        return self._concretize_addr(
            super().concretize_write_addr,
            addr,
            strategies=strategies,
            condition=condition,
        )

    # TUI handler methods.  Fear the boilerplate.
    def divergence_fork(self, addr=None, exprs=None, **kwargs):
        log.warn("Rewinding and forking to avoid conditional dereference")
        raise DivergentAddressSignal(self.state, addr, exprs)

    def divergence_choose(**kwargs):
        raise NotImplementedError(
            "Chosing a specific evaluation is not yet implemented"
        )

    def divergence_ignore(self, **kwargs):
        log.warn("Accepting conditional dereference")

    def divergence_details(self, addr=None, exprs=None, **kwargs):
        log.warn(f"Details of dereference at {self.state.ip}:")
        print_state(log.warn, self.state, "conditional dereference")
        log.warn(f"Address: {addr}")
        log.warn("Possible Evaluations:")
        for expr, result in exprs.items():
            log.warn(f"\t{expr}: {result:x}")

    def divergence_stop(self, **kwargs):
        log.warn("Killing execution path")
        raise PathTerminationSignal()

    def divergence_quit(self, **kwargs):
        log.warn("Aborting execution")
        quit()


class DivergenceExplorationMixin:
    """Mixin for handling exploration-side address concretization

    A memory mixin can detect problematic address concretizations,
    but it can't do anything about them beyond modifying
    a single state.

    This exploration strategy mixin allows the memory mixin
    to handle exceptional cases that need to fork a state.
    """

    def step_state(self, simgr, state, **kwargs):
        try:
            out = super().step_state(simgr, state, **kwargs)
            log.debug(f"Clean Successors: {state} -> {out}")
            for stash, states in out.items():
                if len(states) > 0:
                    if stash is None:
                        stash = "active"
                    for state in states:
                        log.debug(f"\t{state} ({stash}): {state.scratch.guard}")

        except DivergentAddressSignal as e:
            # Fault: we got a divergent address.
            #
            # This means that we tried to dereference a
            # conditional symbolic expression.
            # Normal address concretization causes a big loss of precision;
            # binding a conditional expression doesn't clearly bind any
            # one symbol.
            #
            # One way to handle this is to fork the state,
            # with one state per possible evaluation of the
            # conditional expression.
            #
            # The possible values and their concretizations
            # are computed in the memory plugin,
            # but it's not possible to fork a state in that plugin.
            # It sends us its results via a DivergentAddressException,
            # and we fix it up here.

            # Set up the successors dict.
            out = {None: list(), "unsat": list()}
            log.debug(f"Divergent successors: {state} ->")

            # Bind all address concretizations
            backup = e.state.copy()
            for expr, addr in e.results.items():
                backup.solver.add(expr == addr)

            # Fork a new state for each possible
            # binding of the conditional address.
            for expr, addr in e.results.items():
                new_state = backup.copy()
                new_state.solver.add(e.addr == addr)

                # Test the new state for satisfiability,
                # just in case the concretizer misses something.
                # I wish I didn't have to do this, since it's slow.
                if new_state.solver.satisfiable():
                    log.debug(f"\t{new_state} (active): {e.addr == addr}")
                    out[None].append(new_state)
                else:
                    out["unsat"].append(new_state)
                    log.debug(f"\t{new_state} (unsat): {e.addr == addr}")
        return out



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import ArmHFSysVModel


class ArmHFSysVBsdSignal(BsdSignal, ArmHFSysVModel):
    pass


class ArmHFSysVKill(Kill, ArmHFSysVModel):
    pass


class ArmHFSysVKillpg(Killpg, ArmHFSysVModel):
    pass


class ArmHFSysVPthreadKill(PthreadKill, ArmHFSysVModel):
    pass


class ArmHFSysVPthreadSigmask(PthreadSigmask, ArmHFSysVModel):
    pass


class ArmHFSysVSigaction(Sigaction, ArmHFSysVModel):
    pass


class ArmHFSysVSigaddset(Sigaddset, ArmHFSysVModel):
    pass


class ArmHFSysVSigaltstack(Sigaltstack, ArmHFSysVModel):
    pass


class ArmHFSysVSigdelset(Sigdelset, ArmHFSysVModel):
    pass


class ArmHFSysVSigemptyset(Sigemptyset, ArmHFSysVModel):
    pass


class ArmHFSysVSigfillset(Sigfillset, ArmHFSysVModel):
    pass


class ArmHFSysVSighold(Sighold, ArmHFSysVModel):
    pass


class ArmHFSysVSigignore(Sigignore, ArmHFSysVModel):
    pass


class ArmHFSysVSiginterrupt(Siginterrupt, ArmHFSysVModel):
    pass


class ArmHFSysVSigismember(Sigismember, ArmHFSysVModel):
    pass


class ArmHFSysVSigpause(Sigpause, ArmHFSysVModel):
    pass


class ArmHFSysVSigpending(Sigpending, ArmHFSysVModel):
    pass


class ArmHFSysVSigprocmask(Sigprocmask, ArmHFSysVModel):
    pass


class ArmHFSysVSigqueue(Sigqueue, ArmHFSysVModel):
    pass


class ArmHFSysVSigrelse(Sigrelse, ArmHFSysVModel):
    pass


class ArmHFSysVSigset(Sigset, ArmHFSysVModel):
    pass


class ArmHFSysVSigsuspend(Sigsuspend, ArmHFSysVModel):
    pass


class ArmHFSysVSigtimedwait(Sigtimedwait, ArmHFSysVModel):
    pass


class ArmHFSysVSigwait(Sigwait, ArmHFSysVModel):
    pass


class ArmHFSysVSigwaitinfo(Sigwaitinfo, ArmHFSysVModel):
    pass


__all__ = [
    "ArmHFSysVBsdSignal",
    "ArmHFSysVKill",
    "ArmHFSysVKillpg",
    "ArmHFSysVPthreadKill",
    "ArmHFSysVPthreadSigmask",
    "ArmHFSysVSigaction",
    "ArmHFSysVSigaddset",
    "ArmHFSysVSigaltstack",
    "ArmHFSysVSigdelset",
    "ArmHFSysVSigemptyset",
    "ArmHFSysVSigfillset",
    "ArmHFSysVSighold",
    "ArmHFSysVSigignore",
    "ArmHFSysVSiginterrupt",
    "ArmHFSysVSigismember",
    "ArmHFSysVSigpause",
    "ArmHFSysVSigpending",
    "ArmHFSysVSigprocmask",
    "ArmHFSysVSigqueue",
    "ArmHFSysVSigrelse",
    "ArmHFSysVSigset",
    "ArmHFSysVSigsuspend",
    "ArmHFSysVSigtimedwait",
    "ArmHFSysVSigwait",
    "ArmHFSysVSigwaitinfo",
]



import logging

import claripy

log = logging.getLogger(__name__)


class ClaripyVisitor:
    """Abstract class for a claripy AST visitor"""

    def visit_int(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing Integer")

    def visit_bvv(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing BVV")

    def visit_bvs(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing BVS")

    def visit_add(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing Add")

    def visit_mul(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing Mul")

    def visit_bwor(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing BWOr")

    def visit_concat(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing Concat")

    def visit_extract(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing Extract")

    def visit_if(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing If")

    def visit_reverse(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing Reverse")

    def visit_lshift(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing LShift")

    def visit_rshift(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing RShift")

    def visit_zeroext(self, v, **kwargs):
        raise NotImplementedError(f"{type(self)} missing ZeroExt")

    def visit(self, v, **kwargs):
        """Perform the visitor pattern on a claripy expression."""
        # TODO: This is very far from complete.
        # claripy supports a dizzying array of bitvector operations.
        # It's easier to populate them as I find them.
        if isinstance(v, int):
            # v is a python integer.
            # I occasionally see an int instead of a BVV.
            # I don't exactly know why.
            return self.visit_int(v, **kwargs)
        elif v.op == "BVV":
            # v is a concrete value
            return self.visit_bvv(v, **kwargs)
        elif v.op == "BVS":
            # v is a symbol
            return self.visit_bvs(v, **kwargs)
        elif v.op == "__add__":
            # v = sum(*X), for two or more expressions x in X.
            return self.visit_add(v, **kwargs)
        elif v.op == "__mul__":
            # v = prod(*X), for two or more expressions x in X.
            return self.visit_mul(v, **kwargs)
        elif v.op == "__or__":
            # v = BWOr(*X) for two or more expressions x in X.
            return self.visit_bwor(v, **kwargs)
        elif v.op == "Concat":
            # v = Concat(*X), for two or more expressions x in X.
            return self.visit_concat(v, **kwargs)
        elif v.op == "If":
            # v is ITE(x, y, z)
            return self.visit_if(v, **kwargs)
        elif v.op == "Reverse":
            # v is Reverse(x)
            return self.visit_reverse(v, **kwargs)
        elif v.op == "Extract":
            # v is x[a:b], for expression x and ints a and b.
            return self.visit_extract(v, **kwargs)
        elif v.op == "__lshift__":
            # v is x << y, for expressions x and y
            return self.visit_lshift(v, **kwargs)
        elif v.op == "LShR":
            # v is x >>> y, for expressions x and y
            return self.visit_rshift(v, **kwargs)
        elif v.op == "ZeroExt":
            # v is ZeroExt(v, a) for expression x and int a
            return self.visit_zeroext(v, **kwargs)
        else:
            # v is something I haven't thought of yet.
            raise NotImplementedError(f"Unknown op {v.op}")


class ConditionalVisitor(ClaripyVisitor):
    """Visitor that splits a conditional expression into its possible evaluations."""

    def visit_int(self, v):
        # Integers are leaf ASTs, and never conditional.
        return [(v, claripy.BoolV(True))]

    def visit_bvv(self, v):
        # BVVs are leaf ASTs, and never conditional.
        return [(v, claripy.BoolV(True))]

    def visit_bvs(self, v):
        # BVSs are leaf ASTs, and never conditional.
        return [(v, claripy.BoolV(True))]

    def visit_add(self, v):
        # Addition can produce all combinations of evaluations
        # of the argument expressions.
        out = self.visit(v.args[0])
        for arg in v.args[1:]:
            old_out = out
            out = list()
            for res2, guard2 in self.visit(arg):
                for res1, guard1 in old_out:
                    res = res1 + res2
                    guard = claripy.And(guard1, guard2)
                    out.append((res, guard))
        return out

    def visit_mul(self, v):
        # Multiplication can produce all combinations of evaluations
        # of the argument expressions.
        out = self.visit(v.args[0])
        for arg in v.args[1:]:
            old_out = out
            out = list()
            for res2, guard2 in self.visit(arg):
                for res1, guard1 in old_out:
                    res = res1 * res2
                    guard = claripy.And(guard1, guard2)
                    out.append((res, guard))
        return out

    def visit_concat(self, v):
        # Concatenation can produce all combinations of evaluations
        # of the argument expressions.
        out = self.visit(v.args[0])
        for arg in v.args[1:]:
            old_out = out
            out = list()
            for res2, guard2 in self.visit(arg):
                for res1, guard1 in old_out:
                    res = res1.concat(res2)
                    guard = claripy.And(guard1, guard2)
                    out.append((res, guard))
        return out

    def visit_extract(self, v):
        # Extraction produces one expression per evaluation
        # of the main argument.  The other two are always ints.
        a = v.args[0]
        b = v.args[1]
        res = self.visit(v.args[2])
        res = list(map(lambda x: (x[0][a:b], x[1]), res))
        return res

    def visit_if(self, v):
        # ITE produces the union of the results of
        # the "then" and "else" expressions.
        # The condition itself is ignored.
        out = list()
        out.extend(
            map(lambda x: (x[0], claripy.And(x[1], v.args[0])), self.visit(v.args[1]))
        )
        out.extend(
            map(
                lambda x: (x[0], claripy.And(x[1], claripy.Not(v.args[0]))),
                self.visit(v.args[2]),
            )
        )
        return out

    def visit_reverse(self, v):
        # Reversal produces one expression per evaluation of the argument.
        return list(map(lambda x: (claripy.Reverse(x[0]), x[1]), self.visit(v.args[0])))


class EvalVisitor(ClaripyVisitor):
    """Visitor that concretizes a symbolic expression.

    Uses a mapping from variables to concrete values.
    This implementation requires a complete mapping;
    it will raise an exception if it encounters
    a symbol for which it does not have a value.
    """

    def visit_int(self, v, bindings=None):
        return v

    def visit_bvv(self, v, bindings=None):
        return v

    def visit_bvs(self, v, bindings=None):
        if v.args[0] not in bindings:
            raise KeyError("Missing binding for {v}")
        return bindings[v.args[0]]

    def visit_add(self, v, bindings=None):
        # Eval of sum is sum of evals
        out = self.visit(v.args[0], bindings=bindings)
        for x in v.args[1:]:
            out += self.visit(x, bindings=bindings)
        return out

    def visit_concat(self, v, bindings=None):
        # Eval of concat is concat of evals
        return claripy.Concat(map(lambda x: self.visit(x, bindings=bindings), v.args))

    def visit_extract(self, v, bindings=None):
        # Extract only hase one BV argument; the range limits are ints.
        a = v.args[0]
        b = v.args[1]
        return self.visit(v.args[2], bindings=bindings)[a:b]

    def visit_if(self, v, bindings=None):
        # Concretize all three args of the ITE expression.
        i = self.visit(v.args[0], bindings=bindings)
        t = self.visit(v.args[1], bindings=bindings)
        e = self.visit(v.args[2], bindings=bindings)
        return claripy.If(i, t, e)

    def visit_reverse(self, v, bindings=None):
        # Reverse is a simple unary; reverse the result from the arg.
        res = claripy.Reverse(self.visit(v.args[0], bindings=bindings))
        return res


class PPrintVisitor(ClaripyVisitor):
    def visit_int(self, v, **kwargs):
        raise NotImplementedError("Why?")

    def visit_bvv(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + str(v))

    def visit_bvs(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + str(v))

    def visit_add(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + "Add:")
        for arg in v.args:
            self.visit(arg, indent=indent + 1, out=out)

    def visit_bwor(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + "BWOr:")
        for arg in v.args:
            self.visit(arg, indent=indent + 1, out=out)

    def visit_concat(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + "Concat:")
        for arg in v.args:
            self.visit(arg, indent=indent + 1, out=out)

    def visit_extract(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + f"Extract [{v.args[0]}:{v.args[1]}]:")
        self.visit(v.args[2], indent=indent + 1, out=out)

    def visit_if(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + "If:")
        self.visit(v.args[0], indent=indent + 1, out=out)
        out("  " * indent + "Then:")
        self.visit(v.args[1], indent=indent + 1, out=out)
        out("  " * indent + "Else:")
        self.visit(v.args[2], indent=indent + 1, out=out)

    def visit_reverse(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + "Reverse:")
        self.visit(v.args[0], indent=indent + 1, out=out)

    def visit_lshift(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + "LShift:")
        self.visit(v.args[0], indent=indent + 1, out=out)
        out("  " * indent + "By:")
        self.visit(v.args[1], indent=indent + 1, out=out)

    def visit_rshift(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + "RShift (Logical):")
        self.visit(v.args[0], indent=indent + 1, out=out)
        out("  " * indent + "By:")
        self.visit(v.args[1], indent=indent + 1, out=out)

    def visit_zeroext(self, v, **kwargs):
        indent = kwargs.get("indent", 0)
        out = kwargs.get("out", print)
        out("  " * indent + f"ZeroExt[{v.args[0]}]:")
        self.visit(v.args[1], indent=indent + 1, out=out)



from ....posix.libgen import Basename, Dirname
from ..systemv import LoongArch64SysVModel


class LoongArch64SysVBasename(Basename, LoongArch64SysVModel):
    pass


class LoongArch64SysVDirname(Dirname, LoongArch64SysVModel):
    pass


__all__ = [
    "LoongArch64SysVBasename",
    "LoongArch64SysVDirname",
]



import logging
import typing

from .. import emulators, hinting
from . import analysis

logger = logging.getLogger(__name__)


class CodeCoverage(analysis.Analysis):
    """A simple analysis that logs jumps, calls, and returns."""

    name = "code-coverage"
    description = ""
    version = "0.0.1"

    def run(self, machine) -> None:
        emulator = emulators.UnicornEmulator(machine.get_platform())
        coverage: typing.Dict[int, int] = {}

        for step in machine.step(emulator):
            cpu = step.get_cpu()
            pc = cpu.pc.get_content()
            if pc in coverage:
                coverage[pc] += 1
            else:
                coverage[pc] = 1

        hint = hinting.CoverageHint(message="Coverage for execution", coverage=coverage)
        self.hinter.send(hint)



from ..exceptions import PathTerminationSignal


class TerminationExplorationMixin:
    """
    Mixin allowing analyses to terminate a single path.

    This allows analyses to raise an exception
    that aborts successor computation cleanly,
    rather than producing an 'error' state.

    NOTE: To be effective, this needs to be at the top
    of the mixin hierarchy
    """

    def step_state(self, simgr, state, **kwargs):
        try:
            out = super().step_state(simgr, state, **kwargs)
        except PathTerminationSignal:
            out = dict()

        return out



import typing

from .. import emulators
from .instructions import MemoryReferenceOperand


class BSIDMemoryReferenceOperand(MemoryReferenceOperand):
    """Memory Operand based on the base-scale-index-displacement pattern."""

    def __init__(
        self,
        base: typing.Optional[str] = None,
        index: typing.Optional[str] = None,
        scale: int = 1,
        offset: int = 0,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.base = base
        self.index = index
        self.scale = scale
        self.offset = offset

    def address(self, emulator: emulators.Emulator) -> int:
        base = 0
        if self.base is not None:
            base = emulator.read_register(self.base)

        index = 0
        if self.index is not None:
            index = emulator.read_register(self.index)

        return base + self.scale * index + self.offset

    def to_json(self) -> dict:
        return {
            "base": self.base,
            "index": self.index,
            "scale": self.scale,
            "offset": self.offset,
        }

    @classmethod
    def from_json(cls, dict):
        if any(k not in dict for k in ("base", "index", "scale", "offset")):
            raise ValueError(f"malformed {cls.__name__}: {dict!r}")

        return cls(**dict)

    def __repr__(self) -> str:
        string = ""

        if self.base:
            string = f"{self.base}"

        if self.index:
            if self.scale:
                string = f"{string}+{self.scale}*{self.index}"
            else:
                string = f"{string}+{self.index}"

        if self.offset:
            string = f"{string}+{self.offset:x}"

        return f"{self.__class__.__name__}({string})"



from .aarch64 import AArch64
from .amd64 import AMD64
from .arm import ARMv5T, ARMv6M, ARMv6MThumb, ARMv7A, ARMv7M, ARMv7R
from .i386 import I386
from .mips import MIPS32BE, MIPS32EL
from .mips64 import MIPS64BE, MIPS64EL
from .ppc import PowerPC32, PowerPC64
from .prstatus import PrStatus

__all__ = [
    "AArch64",
    "AMD64",
    "ARMv5T",
    "ARMv6M",
    "ARMv6MThumb",
    "ARMv7M",
    "ARMv7R",
    "ARMv7A",
    "I386",
    "MIPS32BE",
    "MIPS32EL",
    "MIPS64BE",
    "MIPS64EL",
    "PowerPC32",
    "PowerPC64",
    "PrStatus",
]



import copy
import json
import logging
from collections.abc import Callable
from dataclasses import asdict, dataclass
from typing import Dict, List, Type

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Hint:
    """Base class for all Hints.

    Arguments:
        message: A message for this Hint.
    """

    message: str
    """A detailed description."""


class Hinter:
    def __init__(self) -> None:
        self.callbacks: Dict[Type[Hint], List[Callable[[Hint], None]]] = {}

    def register(self, clazz: Type[Hint], callback: Callable[[Hint], None]):
        if clazz not in self.callbacks:
            self.callbacks[clazz] = []
        self.callbacks[clazz].append(callback)

    def send(self, hint: Hint) -> None:
        try:
            logger.debug(json.dumps(asdict(hint)))
        except:
            logger.debug(hint.to_json())  # type: ignore

        clazz = hint.__class__
        if clazz in self.callbacks:
            for callback in self.callbacks[clazz]:
                callback(copy.deepcopy(hint))


__all__ = ["Hint", "Hinter"]



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import MIPS64ELSysVModel


class MIPS64ELSysVFclose(Fclose, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFeof(Feof, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFerror(Ferror, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFgetc(Fgetc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFgets(Fgets, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFopen(Fopen, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFprintf(Fprintf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFputc(Fputc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFputs(Fputs, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFread(Fread, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFscanf(Fscanf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFseek(Fseek, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFtell(Ftell, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFwrite(Fwrite, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVGetc(Getc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVGetchar(Getchar, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVGets(Gets, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVPrintf(Printf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVPutc(Putc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVPutchar(Putchar, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVPuts(Puts, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVRemove(Remove, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVRename(Rename, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVRewind(Rewind, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVScanf(Scanf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSnprintf(Snprintf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSprintf(Sprintf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSscanf(Sscanf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVClearerr(Clearerr, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFflush(Fflush, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFreopen(Freopen, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFgetpos(Fgetpos, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFsetpos(Fsetpos, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVTmpfile(Tmpfile, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVTmpnam(Tmpnam, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVUngetc(Ungetc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVVfprintf(Vfprintf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVVfscanf(Vfscanf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVVprintf(Vprintf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVVscanf(Vscanf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVVsnprintf(Vsnprintf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVVsprintf(Vsprintf, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVVsscanf(Vsscanf, MIPS64ELSysVModel):
    pass


__all__ = [
    "MIPS64ELSysVFclose",
    "MIPS64ELSysVFeof",
    "MIPS64ELSysVFerror",
    "MIPS64ELSysVFgetc",
    "MIPS64ELSysVFgets",
    "MIPS64ELSysVFopen",
    "MIPS64ELSysVFprintf",
    "MIPS64ELSysVFputc",
    "MIPS64ELSysVFputs",
    "MIPS64ELSysVFread",
    "MIPS64ELSysVFscanf",
    "MIPS64ELSysVFseek",
    "MIPS64ELSysVFtell",
    "MIPS64ELSysVFwrite",
    "MIPS64ELSysVGetc",
    "MIPS64ELSysVGetchar",
    "MIPS64ELSysVGets",
    "MIPS64ELSysVPrintf",
    "MIPS64ELSysVPutc",
    "MIPS64ELSysVPutchar",
    "MIPS64ELSysVPuts",
    "MIPS64ELSysVRemove",
    "MIPS64ELSysVRename",
    "MIPS64ELSysVRewind",
    "MIPS64ELSysVScanf",
    "MIPS64ELSysVSnprintf",
    "MIPS64ELSysVSprintf",
    "MIPS64ELSysVSscanf",
    "MIPS64ELSysVClearerr",
    "MIPS64ELSysVFflush",
    "MIPS64ELSysVFreopen",
    "MIPS64ELSysVFgetpos",
    "MIPS64ELSysVFsetpos",
    "MIPS64ELSysVTmpfile",
    "MIPS64ELSysVTmpnam",
    "MIPS64ELSysVUngetc",
    "MIPS64ELSysVVfprintf",
    "MIPS64ELSysVVfscanf",
    "MIPS64ELSysVVprintf",
    "MIPS64ELSysVVscanf",
    "MIPS64ELSysVVsnprintf",
    "MIPS64ELSysVVsprintf",
    "MIPS64ELSysVVsscanf",
]



import unicorn

from ....platforms import Architecture, Byteorder
from .machdef import UnicornMachineDef


class i386MachineDef(UnicornMachineDef):
    """Unicorn machine definition for i386"""

    arch = Architecture.X86_32
    byteorder = Byteorder.LITTLE

    uc_arch = unicorn.UC_ARCH_X86
    uc_mode = unicorn.UC_MODE_32

    _registers = {
        # *** General Purpose Registers ***
        "eax": unicorn.x86_const.UC_X86_REG_EAX,
        "ax": unicorn.x86_const.UC_X86_REG_AX,
        "al": unicorn.x86_const.UC_X86_REG_AL,
        "ah": unicorn.x86_const.UC_X86_REG_AH,
        "ebx": unicorn.x86_const.UC_X86_REG_EBX,
        "bx": unicorn.x86_const.UC_X86_REG_BX,
        "bl": unicorn.x86_const.UC_X86_REG_BL,
        "bh": unicorn.x86_const.UC_X86_REG_BH,
        "ecx": unicorn.x86_const.UC_X86_REG_ECX,
        "cx": unicorn.x86_const.UC_X86_REG_CX,
        "cl": unicorn.x86_const.UC_X86_REG_CL,
        "ch": unicorn.x86_const.UC_X86_REG_CH,
        "edx": unicorn.x86_const.UC_X86_REG_EDX,
        "dx": unicorn.x86_const.UC_X86_REG_DX,
        "dl": unicorn.x86_const.UC_X86_REG_DL,
        "dh": unicorn.x86_const.UC_X86_REG_DH,
        "esi": unicorn.x86_const.UC_X86_REG_ESI,
        "si": unicorn.x86_const.UC_X86_REG_SI,
        "sil": unicorn.x86_const.UC_X86_REG_SIL,
        "edi": unicorn.x86_const.UC_X86_REG_EDI,
        "di": unicorn.x86_const.UC_X86_REG_DI,
        "dil": unicorn.x86_const.UC_X86_REG_DIL,
        "ebp": unicorn.x86_const.UC_X86_REG_EBP,
        "bp": unicorn.x86_const.UC_X86_REG_BP,
        "bpl": unicorn.x86_const.UC_X86_REG_BPL,
        "esp": unicorn.x86_const.UC_X86_REG_ESP,
        "sp": unicorn.x86_const.UC_X86_REG_SP,
        "spl": unicorn.x86_const.UC_X86_REG_SPL,
        # *** Instruction Pointer ***
        "eip": unicorn.x86_const.UC_X86_REG_EIP,
        "ip": unicorn.x86_const.UC_X86_REG_IP,
        # *** Segment Registers ***
        "cs": unicorn.x86_const.UC_X86_REG_CS,
        "ss": unicorn.x86_const.UC_X86_REG_SS,
        "ds": unicorn.x86_const.UC_X86_REG_DS,
        "es": unicorn.x86_const.UC_X86_REG_ES,
        "fs": unicorn.x86_const.UC_X86_REG_FS,
        "gs": unicorn.x86_const.UC_X86_REG_GS,
        # *** Flags Registers ***
        "eflags": unicorn.x86_const.UC_X86_REG_EFLAGS,
        "flags": unicorn.x86_const.UC_X86_REG_FLAGS,
        # *** Control Registers ***
        "cr0": unicorn.x86_const.UC_X86_REG_CR0,
        "cr1": unicorn.x86_const.UC_X86_REG_CR1,
        "cr2": unicorn.x86_const.UC_X86_REG_CR2,
        "cr3": unicorn.x86_const.UC_X86_REG_CR3,
        "cr4": unicorn.x86_const.UC_X86_REG_CR4,
        # NOTE: I've got conflicting reports whether cr8 exists in i386.
        "cr8": unicorn.x86_const.UC_X86_REG_INVALID,
        # *** Debug Registers ***
        "dr0": unicorn.x86_const.UC_X86_REG_DR0,
        "dr1": unicorn.x86_const.UC_X86_REG_DR1,
        "dr2": unicorn.x86_const.UC_X86_REG_DR2,
        "dr3": unicorn.x86_const.UC_X86_REG_DR3,
        "dr6": unicorn.x86_const.UC_X86_REG_DR6,
        "dr7": unicorn.x86_const.UC_X86_REG_DR7,
        # *** Descriptor Table Registers
        # NOTE: Yes, this is 6 bytes; 2 byte segment selector plus 4 byte offset
        "gdtr": unicorn.x86_const.UC_X86_REG_GDTR,
        "idtr": unicorn.x86_const.UC_X86_REG_IDTR,
        "ldtr": unicorn.x86_const.UC_X86_REG_LDTR,
        # *** Task Register ***
        # NOTE: Yes, this is 6 bytes; 2 byte segment selector plus 4 byte offset
        "tr": unicorn.x86_const.UC_X86_REG_TR,
        # *** x87 registers ***
        # NOTE: x87 is supported by Unicorn, but not by SmallWorld.
        # Values are represented as tuples (exponent: int, mantissa: int).
        # If you need x87 support, open a ticket.
        "fpr0": unicorn.x86_const.UC_X86_REG_INVALID,
        "fpr1": unicorn.x86_const.UC_X86_REG_INVALID,
        "fpr2": unicorn.x86_const.UC_X86_REG_INVALID,
        "fpr3": unicorn.x86_const.UC_X86_REG_INVALID,
        "fpr4": unicorn.x86_const.UC_X86_REG_INVALID,
        "fpr5": unicorn.x86_const.UC_X86_REG_INVALID,
        "fpr6": unicorn.x86_const.UC_X86_REG_INVALID,
        "fpr7": unicorn.x86_const.UC_X86_REG_INVALID,
        # x87 Control Register
        "fctrl": unicorn.x86_const.UC_X86_REG_FPCW,
        # x87 Status Register
        "fstat": unicorn.x86_const.UC_X86_REG_FPSW,
        # x87 Tag Register
        "ftag": unicorn.x86_const.UC_X86_REG_FPTAG,
        # x87 Last Instruction Register
        "fip": unicorn.x86_const.UC_X86_REG_FIP,
        # x87 Last Operand Pointer
        "fdp": unicorn.x86_const.UC_X86_REG_FDP,
        # x87 Last Opcode
        "fop": unicorn.x86_const.UC_X86_REG_FOP,
        # NOTE: Docs disagree on the format of fip and fdp.
        # One source describes them as 48-bit offset-plus-segment,
        # the other describes them as 64-bit.
        # There may also be separate segment registers.
        # If you care about the x87 debug info, please feel free to update.
        # *** MMX Registers ***
        # NOTE: The MMX registers are aliases for the low 8 bytes of the x87 registers.
        # The two subsystems cannot be used simultaneously.
        "mm0": unicorn.x86_const.UC_X86_REG_MM0,
        "mm1": unicorn.x86_const.UC_X86_REG_MM1,
        "mm2": unicorn.x86_const.UC_X86_REG_MM2,
        "mm3": unicorn.x86_const.UC_X86_REG_MM3,
        "mm4": unicorn.x86_const.UC_X86_REG_MM4,
        "mm5": unicorn.x86_const.UC_X86_REG_MM5,
        "mm6": unicorn.x86_const.UC_X86_REG_MM6,
        "mm7": unicorn.x86_const.UC_X86_REG_MM7,
        # *** SSE Registers ***
        "xmm0": unicorn.x86_const.UC_X86_REG_XMM0,
        "xmm1": unicorn.x86_const.UC_X86_REG_XMM1,
        "xmm2": unicorn.x86_const.UC_X86_REG_XMM2,
        "xmm3": unicorn.x86_const.UC_X86_REG_XMM3,
        "xmm4": unicorn.x86_const.UC_X86_REG_XMM4,
        "xmm5": unicorn.x86_const.UC_X86_REG_XMM5,
        "xmm6": unicorn.x86_const.UC_X86_REG_XMM6,
        "xmm7": unicorn.x86_const.UC_X86_REG_XMM7,
    }



from ....platforms import Architecture, Byteorder
from .machdef import GhidraMachineDef


class RISCV64MachineDef(GhidraMachineDef):
    arch = Architecture.RISCV64
    byteorder = Byteorder.LITTLE
    language_id = "RISCV:LE:64:default"

    _registers = {
        # *** General-Purpose Registers ***
        # x0 is wired to 0, and aliased as "zero"
        "x0": "zero",
        "zero": "zero",
        # x1 acts as the link register
        # NOTE:ra is the official name; lr might be an angr invention.
        "x1": "ra",
        "ra": "ra",
        # x2 acts as the stack pointer
        "x2": "sp",
        "sp": "sp",
        # x3 acts as the global pointer
        "x3": "gp",
        "gp": "gp",
        # x4 acts as the thread pointer
        "x4": "tp",
        "tp": "tp",
        # x5 is a temporary register
        "x5": "t0",
        "t0": "t0",
        # x6 is a temporary register
        "x6": "t1",
        "t1": "t1",
        # x7 is a temporary register
        "x7": "t2",
        "t2": "t2",
        # x8 is a callee-saved register
        "x8": "s0",
        "s0": "s0",
        # x9 is a callee-saved register
        "x9": "s1",
        "s1": "s1",
        # x10 is argument 0
        "x10": "a0",
        "a0": "a0",
        # x11 is argument 1
        "x11": "a1",
        "a1": "a1",
        # x12 is argument 2
        "x12": "a2",
        "a2": "a2",
        # x13 is argument 3
        "x13": "a3",
        "a3": "a3",
        # x14 is argument 4
        "x14": "a4",
        "a4": "a4",
        # x15 is argument 5
        "x15": "a5",
        "a5": "a5",
        # x16 is argument 6
        "x16": "a6",
        "a6": "a6",
        # x17 is argument 7
        "x17": "a7",
        "a7": "a7",
        # x18 is a callee-saved register
        "x18": "s2",
        "s2": "s2",
        # x19 is a callee-saved register
        "x19": "s3",
        "s3": "s3",
        # x20 is a callee-saved register
        "x20": "s4",
        "s4": "s4",
        # x21 is a callee-saved register
        "x21": "s5",
        "s5": "s5",
        # x22 is a callee-saved register
        "x22": "s6",
        "s6": "s6",
        # x23 is a callee-saved register
        "x23": "s7",
        "s7": "s7",
        # x24 is a callee-saved register
        "x24": "s8",
        "s8": "s8",
        # x25 is a callee-saved register
        "x25": "s9",
        "s9": "s9",
        # x26 is a callee-saved register
        "x26": "s10",
        "s10": "s10",
        # x27 is a callee-saved register
        "x27": "s11",
        "s11": "s11",
        # x28 is a temporary register
        "x28": "t3",
        "t3": "t3",
        # x29 is a temporary register
        "x29": "t4",
        "t4": "t4",
        # x30 is a temporary register
        "x30": "t5",
        "t5": "t5",
        # x31 is a temporary register
        "x31": "t6",
        "t6": "t6",
        # *** Program Counter ***
        "pc": "pc",
        # *** Floating-Point Registers ***
        # f0 is a temporary register
        "f0": "ft0",
        "ft0": "ft0",
        # f1 is a temporary register
        "f1": "ft1",
        "ft1": "ft1",
        # f2 is a temporary register
        "f2": "ft2",
        "ft2": "ft2",
        # f3 is a temporary register
        "f3": "ft3",
        "ft3": "ft3",
        # f4 is a temporary register
        "f4": "ft4",
        "ft4": "ft4",
        # f5 is a temporary register
        "f5": "ft5",
        "ft5": "ft5",
        # f6 is a temporary register
        "f6": "ft6",
        "ft6": "ft6",
        # f7 is a temporary register
        "f7": "ft7",
        "ft7": "ft7",
        # f8 is a callee saved register
        "f8": "fs0",
        "fs0": "fs0",
        # f9 is a callee saved register
        "f9": "fs1",
        "fs1": "fs1",
        # f10 is argument 0
        "f10": "fa0",
        "fa0": "fa0",
        # f11 is argument 1
        "f11": "fa1",
        "fa1": "fa1",
        # f12 is argument 2
        "f12": "fa2",
        "fa2": "fa2",
        # f13 is argument 3
        "f13": "fa3",
        "fa3": "fa3",
        # f14 is argument 4
        "f14": "fa4",
        "fa4": "fa4",
        # f15 is argument 5
        "f15": "fa5",
        "fa5": "fa5",
        # f16 is argument 6
        "f16": "fa6",
        "fa6": "fa6",
        # f7 is argument 7
        "f17": "fa7",
        "fa7": "fa7",
        # f18 is a callee-saved register
        "f18": "fs2",
        "fs2": "fs2",
        # f19 is a callee-saved register
        "f19": "fs3",
        "fs3": "fs3",
        # f20 is a callee-saved register
        "f20": "fs4",
        "fs4": "fs4",
        # f21 is a callee-saved register
        "f21": "fs5",
        "fs5": "fs5",
        # f22 is a callee-saved register
        "f22": "fs6",
        "fs6": "fs6",
        # f23 is a callee-saved register
        "f23": "fs7",
        "fs7": "fs7",
        # f24 is a callee-saved register
        "f24": "fs8",
        "fs8": "fs8",
        # f25 is a callee-saved register
        "f25": "fs9",
        "fs9": "fs9",
        # f26 is a callee-saved register
        "f26": "fs10",
        "fs10": "fs10",
        # f27 is a callee-saved register
        "f27": "fs11",
        "fs11": "fs11",
        # f28 is a temporary register
        "f28": "ft8",
        "ft8": "ft8",
        # f29 is a temporary register
        "f29": "ft9",
        "ft9": "ft9",
        # f30 is a temporary register
        "f30": "ft10",
        "ft10": "ft10",
        # f31 is a temporary register
        "f31": "ft11",
        "ft11": "ft11",
    }



import copy
import logging

from ... import emulators, exceptions, hinting, instructions, state
from .. import analysis

logger = logging.getLogger(__name__)


class ControlFlowTracer(analysis.Analysis):
    """A simple analysis that logs jumps, calls, and returns.

    Arguments:
        num_instructions: The number of instructions to execute.
    """

    def __init__(self, *args, num_instructions: int = 10, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_instructions = num_instructions

    name = "control-flow-tracer"
    description = ""
    version = "0.0.1"

    def run(self, state: state.Machine) -> None:
        machine = copy.deepcopy(state)
        cpu = machine.get_cpu()
        emulator = emulators.UnicornEmulator(cpu.platform)
        machine.apply(emulator)

        from_instruction = None

        for i in range(self.num_instructions):
            instruction = emulator.current_instruction()
            if from_instruction:
                hint = hinting.ControlFlowHint(
                    message="Control Flow Change",
                    from_instruction=instructions.Instruction.from_capstone(
                        from_instruction
                    ),
                    to_instruction=instructions.Instruction.from_capstone(instruction),
                )
                self.hinter.send(hint)
                from_instruction = None
            if self.is_cfi(instruction):
                from_instruction = instruction
            try:
                emulator.step()
            except exceptions.EmulationStop:
                break
            except exceptions.EmulationError as e:
                exhint = hinting.EmulationException(
                    message="Emulation single step raised an exception",
                    pc=instruction.address,
                    instruction_num=i,
                    exception=str(e),
                )
                self.hinter.send(exhint)
                break

    def is_cfi(self, instruction):
        for g in instruction.groups:
            group_name = instruction.group_name(g)
            if group_name == "jump":
                return True
            elif group_name == "call":
                return True
            elif group_name == "ret":
                return True
        return False



import abc
import logging
import random

from ... import emulators
from .model import Model

logger = logging.getLogger(__name__)

######################################################
# This stuff is private. Dont' add it to __all__     #
######################################################


MAX_STRLEN = 0x10000


# obtain addr of emulator heap memory of this size
# NB: this will map new pages into emulator as needed
def _emu_alloc(emulator: emulators.Emulator, size: int) -> int:
    address = 0x10000
    memory_map = emulator.get_memory_map()
    if len(memory_map) > 0:
        address = memory_map[-1][1]
    emulator.map_memory(address, size)
    return address


def _emu_calloc(emulator: emulators.Emulator, size: int) -> int:
    addr = _emu_alloc(emulator, size)
    emulator.write_memory(addr, b"\0" * size)
    return addr


def _emu_strlen_n(emulator: emulators.Emulator, addr: int, n: int) -> int:
    sl = 0
    while sl <= n:
        b_opt = emulator.read_memory(addr + sl, 1)
        if b_opt is not None:
            b = b_opt[0]
            if b == 0:
                break
            sl += 1
        else:
            assert b_opt is not None
    return sl


def _emu_strlen(emulator: emulators.Emulator, addr: int) -> int:
    return _emu_strlen_n(emulator, addr, MAX_STRLEN)


def _emu_memcpy(emulator: emulators.Emulator, dst: int, src: int, n: int) -> None:
    src_bytes = emulator.read_memory(src, n)
    emulator.write_memory(dst, src_bytes)


def _emu_strncpy(
    emulator: emulators.Emulator, dst: int, src: int, n: int, is_strncpy: bool
) -> None:
    if n == 0:
        return
    if emulator.read_memory(src, 1) is None:
        logger.debug("MEM not available in strncpy read @ {src:x}")
    elif emulator.read_memory(dst, 1) is None:
        logger.debug("MEM not available in strncpy write @ {dst:x}")
    else:
        # at least a byte is available at both src and dst
        # find length of src (not to exceed l)
        l2 = _emu_strlen_n(emulator, src, n)
        l3 = min(n, l2)
        if l3 > 0:
            # read the src string and copy to dst
            src_bytes = emulator.read_memory(src, l3)
            emulator.write_memory(dst, src_bytes)
            if is_strncpy and l3 < n:
                # if src string is less than n bytes then, according to man page
                # strncpy copies 0s to get to n bytes
                emulator.write_memory(dst + l3, b"\0" * (n - l3))


def _emu_strncat(emulator: emulators.Emulator, dst: int, src: int, n: int) -> None:
    if n == 0:
        return
    if emulator.read_memory(src, 1) is None:
        logger.debug("MEM not available in strncpy read @ {src:x}")
    elif emulator.read_memory(dst, 1) is None:
        logger.debug("MEM not available in strncpy write @ {dst:x}")
    else:
        # at least a byte is available at both src and dst
        ld = _emu_strlen_n(emulator, dst, MAX_STRLEN)
        ls = _emu_strlen_n(emulator, src, MAX_STRLEN)
        lsn = min(ls, n)
        b_opt = emulator.read_memory(src, lsn)
        if b_opt is not None:
            src_bytes = b_opt + b"\0"
            emulator.write_memory(dst + ld, src_bytes)
        else:
            assert b_opt is not None


######################################################
#                 End private stuff                  #
######################################################


class PosixModel(Model):
    # Base class for POSIX models.
    #
    # Most POSIX ABIS use a register-based calling convention.
    # Rather than tailor each model for each ABI, this base class
    # lets us play madlibs with the argument and return registers.
    #
    # TODO: i386 and mips32 also have stack-passed arguments.

    @property
    @abc.abstractmethod
    def argument1(self) -> str:
        return ""

    @property
    @abc.abstractmethod
    def argument2(self) -> str:
        return ""

    @property
    @abc.abstractmethod
    def argument3(self) -> str:
        return ""

    @property
    @abc.abstractmethod
    def argument4(self) -> str:
        return ""

    @property
    @abc.abstractmethod
    def argument5(self) -> str:
        return ""

    @property
    @abc.abstractmethod
    def argument6(self) -> str:
        return ""

    @property
    @abc.abstractmethod
    def return_val(self) -> str:
        return ""


class BasenameModel(PosixModel):
    name = "basename"

    def model(self, emulator: emulators.Emulator) -> None:
        # char *basename(char "path)
        # returns final component (after last '/') of filename
        # if path ends with '/' return a ptr to an empty string
        path = emulator.read_register(self.argument1)
        sl = _emu_strlen_n(emulator, path, MAX_STRLEN)
        bn = None
        while True:
            b_opt = emulator.read_memory(path + sl, 1)
            if b_opt is not None:
                b = b_opt[0]
                if chr(b) == "/":
                    bn = path + sl + 1
                    break
                if sl == 0:
                    bn = path
                    break
                sl = sl - 1
            else:
                assert b_opt is not None
        assert not (bn is None)
        emulator.write_register(self.return_val, bn)


class CallocModel(PosixModel):
    name = "calloc"

    def model(self, emulator: emulators.Emulator) -> None:
        # void *calloc(size_t count, size_t size);
        count = emulator.read_register(self.argument1)
        size = emulator.read_register(self.argument2)
        num_bytes = count * size
        addr = _emu_calloc(emulator, num_bytes)
        emulator.write_register(self.return_val, addr)


class ReturnsNothingModel(PosixModel):
    def model(self, emulator: emulators.Emulator) -> None:
        # so just do nothing
        pass


class Returns0Model(PosixModel):
    def model(self, emulator: emulators.Emulator) -> None:
        # just return 0
        emulator.write_register(self.return_val, 0)


class DaemonModel(Returns0Model):
    name = "daemon"


class FlockModel(Returns0Model):
    name = "flock"


class Getopt_longModel(PosixModel):
    name = "getopt_long"

    def model(self, emulator: emulators.Emulator) -> None:
        # int getopt_long(int argc, char * const *argv, const char *optstring, const struct option *longopts, int *longindex);
        # return -1 if no more args
        emulator.write_register(self.return_val, -1)


class GetpagesizeModel(PosixModel):
    name = "getpagesize"

    def model(self, emulator: emulators.Emulator) -> None:
        # int getpagesize(void);
        emulator.write_register(self.return_val, 0x1000)


class GetppidModel(Returns0Model):
    name = "getppid"


class GetsModel(PosixModel):
    name = "gets"

    def model(self, emulator: emulators.Emulator) -> None:
        # char *gets(char *str);
        s = emulator.read_register(self.argument1)
        value = input()
        emulator.write_memory(s, value.encode("utf-8"))
        emulator.write_register(self.return_val, s)


class MallocModel(CallocModel):
    name = "malloc"


class MemcpyModel(PosixModel):
    name = "memcpy"

    def model(self, emulator: emulators.Emulator) -> None:
        # void *memcpy(void *restrict dst, const void *restrict src, size_t n);
        dst = emulator.read_register(self.argument1)
        src = emulator.read_register(self.argument2)
        n = emulator.read_register(self.argument3)
        _emu_memcpy(emulator, dst, src, n)
        emulator.write_register(self.return_val, dst)


class OpenModel(PosixModel):
    name = "open"

    def model(self, emulator: emulators.Emulator) -> None:
        # int open64(const char *pathname, int oflag,...);
        # just return a fd that's not stdin/stdout/stderr
        emulator.write_register(self.return_val, 3)


class Open64Model(OpenModel):
    name = "open64"


class PutsModel(PosixModel):
    name = "puts"

    def model(self, emulator: emulators.Emulator) -> None:
        # int fputs(const char *restrict s, FILE *restrict stream);
        addr = emulator.read_register_content(self.argument1)
        sl = _emu_strlen(emulator, addr)
        b_opt = emulator.read_memory(addr, sl)
        if b_opt is not None:
            logger.debug(f"puts stream={self.argument2:x} s=[{b_opt!r}]")
        emulator.write_register(self.return_val, 0)


class PthreadCondInitModel(Returns0Model):
    name = "pthread_cond_init"


class PthreadCondSignalModel(Returns0Model):
    name = "pthread_cond_signal_model"


class PthreadCondWaitModel(Returns0Model):
    name = "pthread_cond_wait"


class PthreadCreateModel(Returns0Model):
    name = "pthread_create"


class PthreadMutexInitModel(Returns0Model):
    name = "pthread_mutex_init"


class PthreadMutexLockModel(Returns0Model):
    name = "ptherad_mutex_lock"


class PthreadMutexUnlockModel(Returns0Model):
    name = "pthread_mutex_unlock"


class PtraceModel(Returns0Model):
    name = "ptrace"


class RandModel(PosixModel):
    name = "rand"

    def model(self, emulator: emulators.Emulator) -> None:
        # int rand(void);
        emulator.write_register(self.return_val, random.randint(0, 0x7FFFFFFF))


class RandomModel(PosixModel):
    name = "random"

    def model(self, emulator: emulators.Emulator) -> None:
        # long random(void);
        emulator.write_register(self.return_val, random.randint(0, 0xFFFFFFFF))


class SleepModel(Returns0Model):
    name = "sleep"


class SrandModel(ReturnsNothingModel):
    name = "srand"


class SrandomModel(ReturnsNothingModel):
    name = "srandom"


class StrcatModel(PosixModel):
    name = "strcat"

    def model(self, emulator: emulators.Emulator) -> None:
        # char *strcat(char *restrict s1, const char *restrict s2);
        dst = emulator.read_register(self.argument1)
        src = emulator.read_register(self.argument2)
        _emu_strncat(emulator, dst, src, MAX_STRLEN)


class StrncatModel(PosixModel):
    name = "strncat"

    def model(self, emulator: emulators.Emulator) -> None:
        # char *strncat(char *restrict s1, const char *restrict s2, size_t n);
        dst = emulator.read_register(self.argument1)
        src = emulator.read_register(self.argument2)
        msl = emulator.read_register(self.argument3)
        _emu_strncat(emulator, dst, src, msl)


class StrcpyModel(PosixModel):
    name = "strcpy"

    def model(self, emulator: emulators.Emulator) -> None:
        # char *strcpy(char * dst, const char * src);
        dst = emulator.read_register(self.argument1)
        src = emulator.read_register(self.argument2)
        _emu_strncpy(emulator, dst, src, MAX_STRLEN, is_strncpy=False)


class StrncpyModel(PosixModel):
    name = "strncpy"

    def model(self, emulator: emulators.Emulator) -> None:
        # char *strncpy(char * dst, const char * src, size_t len);
        dst = emulator.read_register(self.argument1)
        src = emulator.read_register(self.argument2)
        sl = emulator.read_register(self.argument3)
        _emu_strncpy(emulator, dst, src, sl, is_strncpy=True)


class StrdupModel(PosixModel):
    name = "strdup"

    def model(self, emulator: emulators.Emulator) -> None:
        # char *strdup(const char *s1);
        src = emulator.read_register(self.argument1)
        sl = _emu_strlen_n(emulator, src, MAX_STRLEN)
        dst = _emu_calloc(emulator, sl)
        _emu_memcpy(emulator, dst, src, sl)
        emulator.write_register(self.return_val, dst)


class StrlenModel(PosixModel):
    name = "strlen"

    def model(self, emulator: emulators.Emulator) -> None:
        # size_t strlen(const char *s)
        ptr = emulator.read_register(self.argument1)
        emulator.write_register(
            self.return_val, _emu_strlen_n(emulator, ptr, MAX_STRLEN)
        )


class StrnlenModel(PosixModel):
    name = "strnlen"

    def model(self, emulator: emulators.Emulator) -> None:
        # size_t strnlen(const char *s, size_t maxlen);
        ptr = emulator.read_register(self.argument1)
        n = emulator.read_register(self.argument2)
        emulator.write_register(self.return_val, _emu_strlen_n(emulator, ptr, n))


class SysconfModel(PosixModel):
    name = "sysconf"

    def model(self, emulator: emulators.Emulator) -> None:
        # long sysconf(int name);
        arg1 = emulator.read_register(self.argument1)
        if arg1 == 0x54:
            # this is the one case we know about so far
            # it's asking for the number of cores
            emulator.write_register(self.return_val, 1)


class TimeModel(PosixModel):
    name = "time"

    def model(self, emulator: emulators.Emulator) -> None:
        # time_t time(time_t *tloc);
        # just return something
        emulator.write_register(self.return_val, 0x1234)


class UnlinkModel(Returns0Model):
    name = "unlink"


class WriteModel(PosixModel):
    name = "write"

    def model(self, emulator: emulators.Emulator) -> None:
        # ssize_t write(int fildes, const void *buf, size_t nbyte);
        addr = emulator.read_register_content(self.argument2)
        size = emulator.read_register_content(self.argument3)
        b_opt = emulator.read_memory(addr, size)
        if b_opt is not None:
            logger.debug(f"write fd={self.argument1} buf=[{b_opt!r}]")
        emulator.write_register(self.return_val, 0)



from .aarch64 import AArch64
from .amd64 import AMD64
from .arm import ARMv5T, ARMv6M, ARMv6MThumb, ARMv7A, ARMv7M, ARMv7R
from .cpu import *  # noqa: F401, F403
from .cpu import __all__ as __cpu__
from .i386 import I386
from .loongarch import LoongArch64
from .mips import MIPSBE, MIPSEL
from .mips64 import MIPS64BE, MIPS64EL
from .powerpc import PowerPC32, PowerPC64
from .riscv import RISCV64
from .xtensa import XTensaBE, XTensaEL

__all__ = __cpu__ + [
    "AArch64",
    "AMD64",
    "ARMv5T",
    "ARMv6M",
    "ARMv6MThumb",
    "ARMv7M",
    "ARMv7R",
    "ARMv7A",
    "I386",
    "LoongArch64",
    "MIPS64EL",
    "MIPS64BE",
    "MIPSEL",
    "MIPSBE",
    "PowerPC32",
    "PowerPC64",
    "RISCV64",
    "XTensaBE",
    "XTensaEL",
]



import typing

import angr
import archinfo

from ....platforms import Architecture, Byteorder
from .machdef import AngrMachineDef

# angr has no default calling convention for RISCV64
# Let's fix that.


class SimCCRISCV64(angr.calling_conventions.SimCC):
    ARG_REGS = ["a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7"]
    FP_ARG_REGS = ["fa0", "fa1", "fa2", "fa3", "fa4", "fa5", "fa6", "fa7"]
    RETURN_VAL = angr.calling_conventions.SimRegArg("a0", 8)
    RETURN_ADDR = angr.calling_conventions.SimRegArg("ra", 8)
    # angr doesn't type this field correctly
    ARCH = archinfo.ArchRISCV64  # type: ignore[assignment]


angr.calling_conventions.register_default_cc("RISCV64", SimCCRISCV64)

# angr ALSO has no default syscall calling convention for RISCV64
# Let's fix that


class SimCCRISCV64LinuxSyscall(angr.calling_conventions.SimCCSyscall):
    # Since the RISCV peeps don't seem to have written their kernel ABI,
    # I RE'd the syscall convention from glibc.
    # It looks like they use the same arg and return regs,
    # except that they repurpose a7 as the syscall number.
    ARG_REGS = ["a0", "a1", "a2", "a3", "a4", "a5", "a6"]
    FP_ARG_REGS: typing.List[str] = []
    RETURN_VAL = angr.calling_conventions.SimRegArg("a0", 8)
    RETURN_ADDR = angr.calling_conventions.SimRegArg("ip_at_syscall", 4)
    # angr doesn't type this field correctly
    ARCH = archinfo.ArchRISCV64  # type: ignore[assignment]

    @classmethod
    def _match(cls, arch, args, sp_data):
        # Never match; only occurs durring syscalls
        return False

    @staticmethod
    def syscall_num(state):
        return state.regs.a7


angr.calling_conventions.register_syscall_cc(
    "RISCV64", "default", SimCCRISCV64LinuxSyscall
)


class RISCV64MachineDef(AngrMachineDef):
    arch = Architecture.RISCV64
    byteorder = Byteorder.LITTLE

    angr_arch = archinfo.ArchRISCV64()

    _registers = {
        # *** General-Purpose Registers ***
        # x0 is wired to 0, and aliased as "zero"
        "x0": "x0",
        "zero": "zero",
        # x1 acts as the link register
        # NOTE: ra is the official name; lr might be an angr invention.
        "x1": "x1",
        "ra": "ra",
        # x2 acts as the stack pointer
        "x2": "x2",
        "sp": "sp",
        # x3 acts as the global pointer
        "x3": "x3",
        "gp": "gp",
        # x4 acts as the thread pointer
        "x4": "x4",
        "tp": "tp",
        # x5 is a temporary register
        "x5": "x5",
        "t0": "t0",
        # x6 is a temporary register
        "x6": "x6",
        "t1": "t1",
        # x7 is a temporary register
        "x7": "x7",
        "t2": "t2",
        # x8 is a callee-saved register
        "x8": "x8",
        "s0": "s0",
        # x9 is a callee-saved register
        "x9": "x9",
        "s1": "s1",
        # x10 is argument 0
        "x10": "x10",
        "a0": "a0",
        # x11 is argument 1
        "x11": "x11",
        "a1": "a1",
        # x12 is argument 2
        "x12": "x12",
        "a2": "a2",
        # x13 is argument 3
        "x13": "x13",
        "a3": "a3",
        # x14 is argument 4
        "x14": "x14",
        "a4": "a4",
        # x15 is argument 5
        "x15": "x15",
        "a5": "a5",
        # x16 is argument 6
        "x16": "x16",
        "a6": "a6",
        # x17 is argument 7
        "x17": "x17",
        "a7": "a7",
        # x18 is a callee-saved register
        "x18": "x18",
        "s2": "s2",
        # x19 is a callee-saved register
        "x19": "x19",
        "s3": "s3",
        # x20 is a callee-saved register
        "x20": "x20",
        "s4": "s4",
        # x21 is a callee-saved register
        "x21": "x21",
        "s5": "s5",
        # x22 is a callee-saved register
        "x22": "x22",
        "s6": "s6",
        # x23 is a callee-saved register
        "x23": "x23",
        "s7": "s7",
        # x24 is a callee-saved register
        "x24": "x24",
        "s8": "s8",
        # x25 is a callee-saved register
        "x25": "x25",
        "s9": "s9",
        # x26 is a callee-saved register
        "x26": "x26",
        "s10": "s10",
        # x27 is a callee-saved register
        "x27": "x27",
        "s11": "s11",
        # x28 is a temporary register
        "x28": "x28",
        "t3": "t3",
        # x29 is a temporary register
        "x29": "x29",
        "t4": "t4",
        # x30 is a temporary register
        "x30": "x30",
        "t5": "t5",
        # x31 is a temporary register
        "x31": "x31",
        "t6": "t6",
        # *** Program Counter ***
        "pc": "pc",
        # *** Floating-Point Registers ***
        # f0 is a temporary register
        "f0": "f0",
        "ft0": "ft0",
        # f1 is a temporary register
        "f1": "f1",
        "ft1": "ft1",
        # f2 is a temporary register
        "f2": "f2",
        "ft2": "ft2",
        # f3 is a temporary register
        "f3": "f3",
        "ft3": "ft3",
        # f4 is a temporary register
        "f4": "f4",
        "ft4": "ft4",
        # f5 is a temporary register
        "f5": "f5",
        "ft5": "ft5",
        # f6 is a temporary register
        "f6": "f6",
        "ft6": "ft6",
        # f7 is a temporary register
        "f7": "f7",
        "ft7": "ft7",
        # f8 is a callee saved register
        "f8": "f8",
        "fs0": "fs0",
        # f9 is a callee saved register
        "f9": "f9",
        "fs1": "fs1",
        # f10 is argument 0
        "f10": "f10",
        "fa0": "fa0",
        # f11 is argument 1
        "f11": "f11",
        "fa1": "fa1",
        # f12 is argument 2
        "f12": "f12",
        "fa2": "fa2",
        # f13 is argument 3
        "f13": "f13",
        "fa3": "fa3",
        # f14 is argument 4
        "f14": "f14",
        "fa4": "fa4",
        # f15 is argument 5
        "f15": "f15",
        "fa5": "fa5",
        # f16 is argument 6
        "f16": "f16",
        "fa6": "fa6",
        # f7 is argument 7
        "f17": "f17",
        "fa7": "fa7",
        # f18 is a callee-saved register
        "f18": "f18",
        "fs2": "fs2",
        # f19 is a callee-saved register
        "f19": "f19",
        "fs3": "fs3",
        # f20 is a callee-saved register
        "f20": "f20",
        "fs4": "fs4",
        # f21 is a callee-saved register
        "f21": "f21",
        "fs5": "fs5",
        # f22 is a callee-saved register
        "f22": "f22",
        "fs6": "fs6",
        # f23 is a callee-saved register
        "f23": "f23",
        "fs7": "fs7",
        # f24 is a callee-saved register
        "f24": "f24",
        "fs8": "fs8",
        # f25 is a callee-saved register
        "f25": "f25",
        "fs9": "fs9",
        # f26 is a callee-saved register
        "f26": "f26",
        "fs10": "fs10",
        # f27 is a callee-saved register
        "f27": "f27",
        "fs11": "fs11",
        # f28 is a temporary register
        "f28": "f28",
        "ft8": "ft8",
        # f29 is a temporary register
        "f29": "f29",
        "ft9": "ft9",
        # f30 is a temporary register
        "f30": "f30",
        "ft10": "ft10",
        # f31 is a temporary register
        "f31": "f31",
        "ft11": "ft11",
    }



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import MIPSELSysVModel


class MIPSELSysVFclose(Fclose, MIPSELSysVModel):
    pass


class MIPSELSysVFeof(Feof, MIPSELSysVModel):
    pass


class MIPSELSysVFerror(Ferror, MIPSELSysVModel):
    pass


class MIPSELSysVFgetc(Fgetc, MIPSELSysVModel):
    pass


class MIPSELSysVFgets(Fgets, MIPSELSysVModel):
    pass


class MIPSELSysVFopen(Fopen, MIPSELSysVModel):
    pass


class MIPSELSysVFprintf(Fprintf, MIPSELSysVModel):
    pass


class MIPSELSysVFputc(Fputc, MIPSELSysVModel):
    pass


class MIPSELSysVFputs(Fputs, MIPSELSysVModel):
    pass


class MIPSELSysVFread(Fread, MIPSELSysVModel):
    pass


class MIPSELSysVFscanf(Fscanf, MIPSELSysVModel):
    pass


class MIPSELSysVFseek(Fseek, MIPSELSysVModel):
    pass


class MIPSELSysVFtell(Ftell, MIPSELSysVModel):
    pass


class MIPSELSysVFwrite(Fwrite, MIPSELSysVModel):
    pass


class MIPSELSysVGetc(Getc, MIPSELSysVModel):
    pass


class MIPSELSysVGetchar(Getchar, MIPSELSysVModel):
    pass


class MIPSELSysVGets(Gets, MIPSELSysVModel):
    pass


class MIPSELSysVPrintf(Printf, MIPSELSysVModel):
    pass


class MIPSELSysVPutc(Putc, MIPSELSysVModel):
    pass


class MIPSELSysVPutchar(Putchar, MIPSELSysVModel):
    pass


class MIPSELSysVPuts(Puts, MIPSELSysVModel):
    pass


class MIPSELSysVRemove(Remove, MIPSELSysVModel):
    pass


class MIPSELSysVRename(Rename, MIPSELSysVModel):
    pass


class MIPSELSysVRewind(Rewind, MIPSELSysVModel):
    pass


class MIPSELSysVScanf(Scanf, MIPSELSysVModel):
    pass


class MIPSELSysVSnprintf(Snprintf, MIPSELSysVModel):
    pass


class MIPSELSysVSprintf(Sprintf, MIPSELSysVModel):
    pass


class MIPSELSysVSscanf(Sscanf, MIPSELSysVModel):
    pass


class MIPSELSysVClearerr(Clearerr, MIPSELSysVModel):
    pass


class MIPSELSysVFflush(Fflush, MIPSELSysVModel):
    pass


class MIPSELSysVFreopen(Freopen, MIPSELSysVModel):
    pass


class MIPSELSysVFgetpos(Fgetpos, MIPSELSysVModel):
    pass


class MIPSELSysVFsetpos(Fsetpos, MIPSELSysVModel):
    pass


class MIPSELSysVTmpfile(Tmpfile, MIPSELSysVModel):
    pass


class MIPSELSysVTmpnam(Tmpnam, MIPSELSysVModel):
    pass


class MIPSELSysVUngetc(Ungetc, MIPSELSysVModel):
    pass


class MIPSELSysVVfprintf(Vfprintf, MIPSELSysVModel):
    pass


class MIPSELSysVVfscanf(Vfscanf, MIPSELSysVModel):
    pass


class MIPSELSysVVprintf(Vprintf, MIPSELSysVModel):
    pass


class MIPSELSysVVscanf(Vscanf, MIPSELSysVModel):
    pass


class MIPSELSysVVsnprintf(Vsnprintf, MIPSELSysVModel):
    pass


class MIPSELSysVVsprintf(Vsprintf, MIPSELSysVModel):
    pass


class MIPSELSysVVsscanf(Vsscanf, MIPSELSysVModel):
    pass


__all__ = [
    "MIPSELSysVFclose",
    "MIPSELSysVFeof",
    "MIPSELSysVFerror",
    "MIPSELSysVFgetc",
    "MIPSELSysVFgets",
    "MIPSELSysVFopen",
    "MIPSELSysVFprintf",
    "MIPSELSysVFputc",
    "MIPSELSysVFputs",
    "MIPSELSysVFread",
    "MIPSELSysVFscanf",
    "MIPSELSysVFseek",
    "MIPSELSysVFtell",
    "MIPSELSysVFwrite",
    "MIPSELSysVGetc",
    "MIPSELSysVGetchar",
    "MIPSELSysVGets",
    "MIPSELSysVPrintf",
    "MIPSELSysVPutc",
    "MIPSELSysVPutchar",
    "MIPSELSysVPuts",
    "MIPSELSysVRemove",
    "MIPSELSysVRename",
    "MIPSELSysVRewind",
    "MIPSELSysVScanf",
    "MIPSELSysVSnprintf",
    "MIPSELSysVSprintf",
    "MIPSELSysVSscanf",
    "MIPSELSysVClearerr",
    "MIPSELSysVFflush",
    "MIPSELSysVFreopen",
    "MIPSELSysVFgetpos",
    "MIPSELSysVFsetpos",
    "MIPSELSysVTmpfile",
    "MIPSELSysVTmpnam",
    "MIPSELSysVUngetc",
    "MIPSELSysVVfprintf",
    "MIPSELSysVVfscanf",
    "MIPSELSysVVprintf",
    "MIPSELSysVVscanf",
    "MIPSELSysVVsnprintf",
    "MIPSELSysVVsprintf",
    "MIPSELSysVVsscanf",
]



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import I386SysVModel


class I386SysVAbs(Abs, I386SysVModel):
    pass


class I386SysVLAbs(LAbs, I386SysVModel):
    pass


class I386SysVLLAbs(LLAbs, I386SysVModel):
    pass


class I386SysVAtof(Atof, I386SysVModel):
    pass


class I386SysVAtoi(Atoi, I386SysVModel):
    pass


class I386SysVAtol(Atol, I386SysVModel):
    pass


class I386SysVAtoll(Atoll, I386SysVModel):
    pass


class I386SysVCalloc(Calloc, I386SysVModel):
    pass


class I386SysVDiv(Div, I386SysVModel):
    pass


class I386SysVLDiv(LDiv, I386SysVModel):
    pass


class I386SysVLLDiv(LLDiv, I386SysVModel):
    pass


class I386SysVExit(Exit, I386SysVModel):
    pass


class I386SysVFree(Free, I386SysVModel):
    pass


class I386SysVMalloc(Malloc, I386SysVModel):
    pass


class I386SysVQSort(QSort, I386SysVModel):
    pass


class I386SysVRand(Rand, I386SysVModel):
    pass


class I386SysVRealloc(Realloc, I386SysVModel):
    pass


class I386SysVSrand(Srand, I386SysVModel):
    pass


class I386SysVAbort(Abort, I386SysVModel):
    pass


class I386SysVAtexit(Atexit, I386SysVModel):
    pass


class I386SysVGetenv(Getenv, I386SysVModel):
    pass


class I386SysVMblen(Mblen, I386SysVModel):
    pass


class I386SysVMbstowcs(Mbstowcs, I386SysVModel):
    pass


class I386SysVMbtowc(Mbtowc, I386SysVModel):
    pass


class I386SysVSystem(System, I386SysVModel):
    pass


class I386SysVWcstombs(Wcstombs, I386SysVModel):
    pass


class I386SysVWctomb(Wctomb, I386SysVModel):
    pass


__all__ = [
    "I386SysVAbs",
    "I386SysVLAbs",
    "I386SysVLLAbs",
    "I386SysVAtof",
    "I386SysVAtoi",
    "I386SysVAtol",
    "I386SysVAtoll",
    "I386SysVCalloc",
    "I386SysVDiv",
    "I386SysVLDiv",
    "I386SysVLLDiv",
    "I386SysVExit",
    "I386SysVFree",
    "I386SysVMalloc",
    "I386SysVQSort",
    "I386SysVRand",
    "I386SysVRealloc",
    "I386SysVSrand",
    "I386SysVAbort",
    "I386SysVAtexit",
    "I386SysVGetenv",
    "I386SysVMblen",
    "I386SysVMbstowcs",
    "I386SysVMbtowc",
    "I386SysVSystem",
    "I386SysVWcstombs",
    "I386SysVWctomb",
]



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import LoongArch64SysVModel


class LoongArch64SysVFclose(Fclose, LoongArch64SysVModel):
    pass


class LoongArch64SysVFeof(Feof, LoongArch64SysVModel):
    pass


class LoongArch64SysVFerror(Ferror, LoongArch64SysVModel):
    pass


class LoongArch64SysVFgetc(Fgetc, LoongArch64SysVModel):
    pass


class LoongArch64SysVFgets(Fgets, LoongArch64SysVModel):
    pass


class LoongArch64SysVFopen(Fopen, LoongArch64SysVModel):
    pass


class LoongArch64SysVFprintf(Fprintf, LoongArch64SysVModel):
    pass


class LoongArch64SysVFputc(Fputc, LoongArch64SysVModel):
    pass


class LoongArch64SysVFputs(Fputs, LoongArch64SysVModel):
    pass


class LoongArch64SysVFread(Fread, LoongArch64SysVModel):
    pass


class LoongArch64SysVFscanf(Fscanf, LoongArch64SysVModel):
    pass


class LoongArch64SysVFseek(Fseek, LoongArch64SysVModel):
    pass


class LoongArch64SysVFtell(Ftell, LoongArch64SysVModel):
    pass


class LoongArch64SysVFwrite(Fwrite, LoongArch64SysVModel):
    pass


class LoongArch64SysVGetc(Getc, LoongArch64SysVModel):
    pass


class LoongArch64SysVGetchar(Getchar, LoongArch64SysVModel):
    pass


class LoongArch64SysVGets(Gets, LoongArch64SysVModel):
    pass


class LoongArch64SysVPrintf(Printf, LoongArch64SysVModel):
    pass


class LoongArch64SysVPutc(Putc, LoongArch64SysVModel):
    pass


class LoongArch64SysVPutchar(Putchar, LoongArch64SysVModel):
    pass


class LoongArch64SysVPuts(Puts, LoongArch64SysVModel):
    pass


class LoongArch64SysVRemove(Remove, LoongArch64SysVModel):
    pass


class LoongArch64SysVRename(Rename, LoongArch64SysVModel):
    pass


class LoongArch64SysVRewind(Rewind, LoongArch64SysVModel):
    pass


class LoongArch64SysVScanf(Scanf, LoongArch64SysVModel):
    pass


class LoongArch64SysVSnprintf(Snprintf, LoongArch64SysVModel):
    pass


class LoongArch64SysVSprintf(Sprintf, LoongArch64SysVModel):
    pass


class LoongArch64SysVSscanf(Sscanf, LoongArch64SysVModel):
    pass


class LoongArch64SysVClearerr(Clearerr, LoongArch64SysVModel):
    pass


class LoongArch64SysVFflush(Fflush, LoongArch64SysVModel):
    pass


class LoongArch64SysVFreopen(Freopen, LoongArch64SysVModel):
    pass


class LoongArch64SysVFgetpos(Fgetpos, LoongArch64SysVModel):
    pass


class LoongArch64SysVFsetpos(Fsetpos, LoongArch64SysVModel):
    pass


class LoongArch64SysVTmpfile(Tmpfile, LoongArch64SysVModel):
    pass


class LoongArch64SysVTmpnam(Tmpnam, LoongArch64SysVModel):
    pass


class LoongArch64SysVUngetc(Ungetc, LoongArch64SysVModel):
    pass


class LoongArch64SysVVfprintf(Vfprintf, LoongArch64SysVModel):
    pass


class LoongArch64SysVVfscanf(Vfscanf, LoongArch64SysVModel):
    pass


class LoongArch64SysVVprintf(Vprintf, LoongArch64SysVModel):
    pass


class LoongArch64SysVVscanf(Vscanf, LoongArch64SysVModel):
    pass


class LoongArch64SysVVsnprintf(Vsnprintf, LoongArch64SysVModel):
    pass


class LoongArch64SysVVsprintf(Vsprintf, LoongArch64SysVModel):
    pass


class LoongArch64SysVVsscanf(Vsscanf, LoongArch64SysVModel):
    pass


__all__ = [
    "LoongArch64SysVFclose",
    "LoongArch64SysVFeof",
    "LoongArch64SysVFerror",
    "LoongArch64SysVFgetc",
    "LoongArch64SysVFgets",
    "LoongArch64SysVFopen",
    "LoongArch64SysVFprintf",
    "LoongArch64SysVFputc",
    "LoongArch64SysVFputs",
    "LoongArch64SysVFread",
    "LoongArch64SysVFscanf",
    "LoongArch64SysVFseek",
    "LoongArch64SysVFtell",
    "LoongArch64SysVFwrite",
    "LoongArch64SysVGetc",
    "LoongArch64SysVGetchar",
    "LoongArch64SysVGets",
    "LoongArch64SysVPrintf",
    "LoongArch64SysVPutc",
    "LoongArch64SysVPutchar",
    "LoongArch64SysVPuts",
    "LoongArch64SysVRemove",
    "LoongArch64SysVRename",
    "LoongArch64SysVRewind",
    "LoongArch64SysVScanf",
    "LoongArch64SysVSnprintf",
    "LoongArch64SysVSprintf",
    "LoongArch64SysVSscanf",
    "LoongArch64SysVClearerr",
    "LoongArch64SysVFflush",
    "LoongArch64SysVFreopen",
    "LoongArch64SysVFgetpos",
    "LoongArch64SysVFsetpos",
    "LoongArch64SysVTmpfile",
    "LoongArch64SysVTmpnam",
    "LoongArch64SysVUngetc",
    "LoongArch64SysVVfprintf",
    "LoongArch64SysVVfscanf",
    "LoongArch64SysVVprintf",
    "LoongArch64SysVVscanf",
    "LoongArch64SysVVsnprintf",
    "LoongArch64SysVVsprintf",
    "LoongArch64SysVVsscanf",
]



import abc
import typing
from dataclasses import dataclass

from ... import utils
from ..platforms import Architecture, Byteorder, Platform


@dataclass(frozen=True)
class RegisterDef:
    name: str
    size: int


@dataclass(frozen=True)
class RegisterAliasDef(RegisterDef):
    parent: str
    offset: int


class PlatformDef(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def architecture(self) -> Architecture:
        """Architecture ID for this platform"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def byteorder(self) -> Byteorder:
        """Byteorder for this platform"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def address_size(self) -> int:
        """Address size in bytes"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def capstone_arch(self) -> int:
        """Capstone Architecture ID"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def capstone_mode(self) -> int:
        """Capstone Mode ID"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def conditional_branch_mnemonics(self) -> typing.Set[str]:
        """Set of conditional branch mnemonics"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def compare_mnemonics(self) -> typing.Set[str]:
        """Set of comparison mnemonics"""
        raise NotImplementedError()

    delay_slot_mnemonics: typing.Set[str] = set()
    """Set of delay slot mnemonics"""

    @property
    @abc.abstractmethod
    def pc_register(self) -> str:
        """Program Counter register name"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def sp_register(self) -> str:
        """Stack Pointer register name"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def general_purpose_registers(self) -> typing.List[str]:
        """List of general-purpose register names"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def registers(self) -> typing.Dict[str, RegisterDef]:
        """Mapping from canonical register names to register definitions"""
        raise NotImplementedError()

    @classmethod
    def for_platform(cls, platform: Platform):
        try:
            return utils.find_subclass(
                cls,
                lambda x: x.architecture == platform.architecture
                and x.byteorder == platform.byteorder,
            )
        except:
            raise ValueError(f"No platform definition for {platform}")



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import AMD64SysVModel


class AMD64SysVFclose(Fclose, AMD64SysVModel):
    pass


class AMD64SysVFeof(Feof, AMD64SysVModel):
    pass


class AMD64SysVFerror(Ferror, AMD64SysVModel):
    pass


class AMD64SysVFgetc(Fgetc, AMD64SysVModel):
    pass


class AMD64SysVFgets(Fgets, AMD64SysVModel):
    pass


class AMD64SysVFopen(Fopen, AMD64SysVModel):
    pass


class AMD64SysVFprintf(Fprintf, AMD64SysVModel):
    pass


class AMD64SysVFputc(Fputc, AMD64SysVModel):
    pass


class AMD64SysVFputs(Fputs, AMD64SysVModel):
    pass


class AMD64SysVFread(Fread, AMD64SysVModel):
    pass


class AMD64SysVFscanf(Fscanf, AMD64SysVModel):
    pass


class AMD64SysVFseek(Fseek, AMD64SysVModel):
    pass


class AMD64SysVFtell(Ftell, AMD64SysVModel):
    pass


class AMD64SysVFwrite(Fwrite, AMD64SysVModel):
    pass


class AMD64SysVGetc(Getc, AMD64SysVModel):
    pass


class AMD64SysVGetchar(Getchar, AMD64SysVModel):
    pass


class AMD64SysVGets(Gets, AMD64SysVModel):
    pass


class AMD64SysVPrintf(Printf, AMD64SysVModel):
    pass


class AMD64SysVPutc(Putc, AMD64SysVModel):
    pass


class AMD64SysVPutchar(Putchar, AMD64SysVModel):
    pass


class AMD64SysVPuts(Puts, AMD64SysVModel):
    pass


class AMD64SysVRemove(Remove, AMD64SysVModel):
    pass


class AMD64SysVRename(Rename, AMD64SysVModel):
    pass


class AMD64SysVRewind(Rewind, AMD64SysVModel):
    pass


class AMD64SysVScanf(Scanf, AMD64SysVModel):
    pass


class AMD64SysVSnprintf(Snprintf, AMD64SysVModel):
    pass


class AMD64SysVSprintf(Sprintf, AMD64SysVModel):
    pass


class AMD64SysVSscanf(Sscanf, AMD64SysVModel):
    pass


class AMD64SysVClearerr(Clearerr, AMD64SysVModel):
    pass


class AMD64SysVFflush(Fflush, AMD64SysVModel):
    pass


class AMD64SysVFreopen(Freopen, AMD64SysVModel):
    pass


class AMD64SysVFgetpos(Fgetpos, AMD64SysVModel):
    pass


class AMD64SysVFsetpos(Fsetpos, AMD64SysVModel):
    pass


class AMD64SysVTmpfile(Tmpfile, AMD64SysVModel):
    pass


class AMD64SysVTmpnam(Tmpnam, AMD64SysVModel):
    pass


class AMD64SysVUngetc(Ungetc, AMD64SysVModel):
    pass


class AMD64SysVVfprintf(Vfprintf, AMD64SysVModel):
    pass


class AMD64SysVVfscanf(Vfscanf, AMD64SysVModel):
    pass


class AMD64SysVVprintf(Vprintf, AMD64SysVModel):
    pass


class AMD64SysVVscanf(Vscanf, AMD64SysVModel):
    pass


class AMD64SysVVsnprintf(Vsnprintf, AMD64SysVModel):
    pass


class AMD64SysVVsprintf(Vsprintf, AMD64SysVModel):
    pass


class AMD64SysVVsscanf(Vsscanf, AMD64SysVModel):
    pass


__all__ = [
    "AMD64SysVFclose",
    "AMD64SysVFeof",
    "AMD64SysVFerror",
    "AMD64SysVFgetc",
    "AMD64SysVFgets",
    "AMD64SysVFopen",
    "AMD64SysVFprintf",
    "AMD64SysVFputc",
    "AMD64SysVFputs",
    "AMD64SysVFread",
    "AMD64SysVFscanf",
    "AMD64SysVFseek",
    "AMD64SysVFtell",
    "AMD64SysVFwrite",
    "AMD64SysVGetc",
    "AMD64SysVGetchar",
    "AMD64SysVGets",
    "AMD64SysVPrintf",
    "AMD64SysVPutc",
    "AMD64SysVPutchar",
    "AMD64SysVPuts",
    "AMD64SysVRemove",
    "AMD64SysVRename",
    "AMD64SysVRewind",
    "AMD64SysVScanf",
    "AMD64SysVSnprintf",
    "AMD64SysVSprintf",
    "AMD64SysVSscanf",
    "AMD64SysVClearerr",
    "AMD64SysVFflush",
    "AMD64SysVFreopen",
    "AMD64SysVFgetpos",
    "AMD64SysVFsetpos",
    "AMD64SysVTmpfile",
    "AMD64SysVTmpnam",
    "AMD64SysVUngetc",
    "AMD64SysVVfprintf",
    "AMD64SysVVfscanf",
    "AMD64SysVVprintf",
    "AMD64SysVVscanf",
    "AMD64SysVVsnprintf",
    "AMD64SysVVsprintf",
    "AMD64SysVVsscanf",
]



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import MIPS64SysVModel


class MIPS64SysVMemcpy(Memcpy, MIPS64SysVModel):
    pass


class MIPS64SysVMemmove(Memmove, MIPS64SysVModel):
    pass


class MIPS64SysVStrcpy(Strcpy, MIPS64SysVModel):
    pass


class MIPS64SysVStrncpy(Strncpy, MIPS64SysVModel):
    pass


class MIPS64SysVStrcat(Strcat, MIPS64SysVModel):
    pass


class MIPS64SysVStrncat(Strncat, MIPS64SysVModel):
    pass


class MIPS64SysVMemcmp(Memcmp, MIPS64SysVModel):
    pass


class MIPS64SysVStrncmp(Strncmp, MIPS64SysVModel):
    pass


class MIPS64SysVStrcmp(Strcmp, MIPS64SysVModel):
    pass


class MIPS64SysVStrcoll(Strcoll, MIPS64SysVModel):
    pass


class MIPS64SysVStrxfrm(Strxfrm, MIPS64SysVModel):
    pass


class MIPS64SysVMemchr(Memchr, MIPS64SysVModel):
    pass


class MIPS64SysVStrchr(Strchr, MIPS64SysVModel):
    pass


class MIPS64SysVStrcspn(Strcspn, MIPS64SysVModel):
    pass


class MIPS64SysVStrpbrk(Strpbrk, MIPS64SysVModel):
    pass


class MIPS64SysVStrrchr(Strrchr, MIPS64SysVModel):
    pass


class MIPS64SysVStrspn(Strspn, MIPS64SysVModel):
    pass


class MIPS64SysVStrstr(Strstr, MIPS64SysVModel):
    pass


class MIPS64SysVStrtok(Strtok, MIPS64SysVModel):
    pass


class MIPS64SysVMemset(Memset, MIPS64SysVModel):
    pass


class MIPS64SysVStrerror(Strerror, MIPS64SysVModel):
    pass


class MIPS64SysVStrlen(Strlen, MIPS64SysVModel):
    pass


__all__ = [
    "MIPS64SysVMemcpy",
    "MIPS64SysVMemmove",
    "MIPS64SysVStrcpy",
    "MIPS64SysVStrncpy",
    "MIPS64SysVStrcat",
    "MIPS64SysVStrncat",
    "MIPS64SysVMemcmp",
    "MIPS64SysVStrncmp",
    "MIPS64SysVStrcmp",
    "MIPS64SysVStrcoll",
    "MIPS64SysVStrxfrm",
    "MIPS64SysVMemchr",
    "MIPS64SysVStrchr",
    "MIPS64SysVStrcspn",
    "MIPS64SysVStrpbrk",
    "MIPS64SysVStrrchr",
    "MIPS64SysVStrspn",
    "MIPS64SysVStrstr",
    "MIPS64SysVStrtok",
    "MIPS64SysVMemset",
    "MIPS64SysVStrerror",
    "MIPS64SysVStrlen",
]



import typing

import lief

from ....exceptions import ConfigurationError
from ....platforms import Architecture, Byteorder, Platform, PlatformDef
from ....utils import RangeCollection
from ...state import BytesValue
from ..code import Executable
from .structs import PEExport, PEImport

# PE32+ machine types
# There are a lot more of these, but I can only test on amd64 and i386
IMAGE_FILE_MACHINE_AMD64 = 0x8664
IMAGE_FILE_MACHINE_I386 = 0x14C

# Section flags
# Currently, the only one we care about is "code".
# Not sure how the others interact with file loading.
IMAGE_SCN_CNT_CODE = 0x20

# Relocation types
# All of these adjust based on the difference between
# the requested and actual load addresses
IMAGE_REL_BASED_ABSOLUTE = 0x0  # No-op
IMAGE_REL_BASED_HIGHLOW = 0x3  # Add 32-bit difference to a 32-bit value
IMAGE_REL_BASED_DIR64 = 0xA  # Add 64-bit difference to a 64-bit value


class PEExecutable(Executable):
    """Executable loaded from a PE32+

    This loads a single PE32+ file into a SmallWorld memory object.
    It performs no relocation or initialization,
    just maps the file into memory as the kernel intended.

    Arguments:
        file: File-like object containing the image
        platform: Optional platform; used for header verification
        ignore_platform: Do not try to ID or verify platform from headers
        user_base: Optional user-specified base address
        page_size: System page size
    """

    def __init__(
        self,
        file: typing.BinaryIO,
        platform: typing.Optional[Platform] = None,
        ignore_platform: bool = False,
        user_base: typing.Optional[int] = None,
        page_size: int = 0x1000,
    ):
        # Initialize with null address and size;
        # we will update these later
        super().__init__(0, 0)

        self.platform = platform
        self.platdef: typing.Optional[PlatformDef] = None
        self.bounds: RangeCollection = RangeCollection()

        self._page_size = page_size
        self._user_base = user_base
        self._file_base = 0

        self._exports: typing.List[PEExport] = list()
        self._exports_by_name: typing.Dict[typing.Tuple[str, str], PEExport] = dict()
        self._exports_by_ordinal: typing.Dict[typing.Tuple[str, int], PEExport] = dict()

        self._imports: typing.List[PEImport] = list()
        self._imports_by_name: typing.Dict[typing.Tuple[str, str], PEImport] = dict()
        self._imports_by_ordinal: typing.Dict[typing.Tuple[str, int], PEImport] = dict()

        # Read the entire image out of the file
        image = file.read()

        # Use lief to check if this is a PE
        # NOTE: For some reason, this takes list(int), not bytes
        if not lief.is_pe(list(image)):
            raise ConfigurationError("Image is not a PE")

        # Use lief to parse the PE
        # NOTE: For some reason, this takes list(int), not bytes
        # NOTE: lief objects aren't deep-copyable.
        # I'd love to keep `pe` around for later use, but I can't.
        pe = lief.PE.parse(list(image))
        if pe is None:
            raise ConfigurationError("Failed parsing PE image")

        # Check machine compatibility
        if not ignore_platform:
            hdr_platform = self._platform_for_chdr(pe)
            if self.platform is not None:
                if self.platform != hdr_platform:
                    raise ConfigurationError(
                        "Platform mismatch: "
                        f"specified {self.platform}, but got {hdr_platform} from header"
                    )
            else:
                self.platform = hdr_platform
            self.platdef = PlatformDef.for_platform(hdr_platform)

        # Determine the file base address
        self._file_base = pe.imagebase
        self._determine_base()

        for section in pe.sections:
            # TODO: How do PE32+ files distinguish loadable segments?
            # mingw objdump has some notion of allocated segments,
            # but it doesn't map to anything in any of the flags.
            self._map_section(section, image)

        # Fix up the memory image to account for changes in the base address
        self._apply_base_relocations(pe)

        # Compute the final total capacity
        for offset, value in self.items():
            self.size = max(self.size, offset + value.get_size())

        # Extract exports
        self._extract_exports(pe)

        # Extract imports
        self._extract_imports(pe)

    def _platform_for_chdr(self, pe):
        if pe.header.machine.value == IMAGE_FILE_MACHINE_AMD64:
            return Platform(Architecture.X86_64, Byteorder.LITTLE)
        elif pe.header.machine.value == IMAGE_FILE_MACHINE_I386:
            return Platform(Architecture.X86_32, Byteorder.LITTLE)
        else:
            raise ConfigurationError(f"Unsupported machine type {pe.header.machine}")

    def _determine_base(self):
        if self._user_base is None:
            # No user base requested
            if self._file_base == 0:
                # No file base defined (unusal for PE32+)
                # Need the user to provide one
                raise ConfigurationError(
                    "No base address provided, and none defined in PE file"
                )
            else:
                self.address = self._file_base
        else:
            # PE files are always position-independent;
            # we'll just have to fix things up later.
            self.address = self._user_base

    def _rebase_file(self, val: int):
        # Rebase an offset from file-relative to image-relative
        res = val + self.address
        return res

    def _page_align(self, val: int, up: bool = True):
        if up:
            val += self._page_size - 1
        return (val // self._page_size) * self._page_size

    def _map_section(self, sect, image):
        # NOTE: Unlike ELFs, The physical address of PE sections is not page-aligned.
        sect_start = sect.offset
        sect_end = sect.offset + sect.size
        sect_addr = self._page_align(self._rebase_file(sect.virtual_address), up=False)
        sect_size = self._page_align(sect.virtual_size)

        sect_data = image[sect_start:sect_end]
        if len(sect_data) < sect_size:
            # Section is shorter than what's available in the file;
            # THis will get zero-padded.
            pad_size = sect_size - len(sect_data)
            sect_data += b"\0" * pad_size
        if len(sect_data) != sect_size:
            raise ConfigurationError(
                f"Expected segment of size {sect_size}, but got {len(sect_data)}"
            )

        if 0 != (sect.characteristics & IMAGE_SCN_CNT_CODE):
            # This is a code segment; add it to program bounds
            self.bounds.add_range((sect_addr, sect_addr + sect_size))

        sect_value = BytesValue(sect_data, None)
        self[sect_addr - self.address] = sect_value

    def _apply_base_relocations(self, pe):
        # PE32+ files always allow the loader to override
        # the program's base address.
        #
        # The relocation section marks places where
        # an absolute offset needs fixing
        # if the loader decides to override
        if self.address == self._file_base:
            # This file was not relocated.  Nothing to do.
            return

        if self.platform.byteorder == Byteorder.LITTLE:
            byteorder = "little"
        elif self.platform.byteorder == Byteorder.BIG:
            byteorder = "big"
        else:
            raise ConfigurationError(
                f"Can't encode byteorder {self.platform.byteorder}"
            )

        for base_reloc in pe.relocations:
            for entry in base_reloc.entries:
                if entry.type.value == IMAGE_REL_BASED_ABSOLUTE:
                    # Entry is a no-op; continue
                    continue
                for off, val in self.items():
                    if entry.address >= off and entry.address <= val.get_size():
                        # Find the section containing this address
                        contents = bytearray(val.get_content())
                        fix_off = entry.address - off

                        if entry.type.value == IMAGE_REL_BASED_HIGHLOW:
                            # 4-byte repair
                            tofix = int.from_bytes(
                                contents[fix_off : fix_off + 4], byteorder
                            )
                            tofix += self.address - self._file_base
                            contents[fix_off : fix_off + 4] = tofix.to_bytes(
                                4, byteorder
                            )

                        elif entry.type.value == IMAGE_REL_BASED_DIR64:
                            # 8-byte repair
                            tofix = int.from_bytes(
                                contents[fix_off : fix_off + 8], byteorder
                            )
                            tofix += self.address - self._file_base
                            contents[fix_off : fix_off + 8] = tofix.to_bytes(
                                8, byteorder
                            )

                        else:
                            raise ConfigurationError(
                                "Unhandled relocation type {entry.type}"
                            )
                        # Reset the data
                        val.set_content(bytes(contents))
                        break

    def _extract_exports(self, pe) -> None:
        if not pe.has_exports:
            return

        d = pe.get_export()
        for e in d.entries:
            if e.is_forwarded:
                raise NotImplementedError(
                    "{d.name}.{e.name} is forwarded to {e.forward_information}"
                )
            else:
                exp = PEExport(
                    dll=d.name,
                    name=e.name,
                    ordinal=e.ordinal,
                    forwarder=None,
                    value=e.value + self.address,
                )
            self._exports.append(exp)
            self._exports_by_name[(d.name, e.name)] = exp
            self._exports_by_ordinal[(d.name, e.ordinal)] = exp

    def _extract_imports(self, pe) -> None:
        if not pe.has_imports:
            return

        # Iterate over the import directories
        # These group imports by the DLL that provides them
        for d in pe.imports:
            for e in d.entries:
                imp = PEImport(
                    dll=d.name,
                    name=e.name if not e.is_ordinal else None,
                    ordinal=e.ordinal if e.is_ordinal else None,
                    iat_address=e.iat_address + self.address,
                    forwarder=None,
                    value=None,
                )
                if e.name == "puts":
                    print(f"puts IAT at {e.iat_address:x} or {e.iat_value:x}")
                self._imports.append(imp)
                if e.is_ordinal:
                    self._imports_by_ordinal[(d.name, e.ordinal)] = imp
                else:
                    self._imports_by_name[(d.name, e.name)] = imp

    def update_import(
        self, dll: str, hint: typing.Union[str, int, PEImport], value: int
    ) -> None:
        """Update an imported symbol in this PE file

        Arguments:
            dll: Name of the DLL where the import is defined
            hint: Symbol name or ordinal
            value: New value of the symbol
        """
        # No platform specified
        if self.platform is None or self.platdef is None:
            raise ConfigurationError("No platform specified; can't update imports")

        if isinstance(hint, str):
            # Look up by name
            if (dll, hint) not in self._imports_by_name:
                raise ConfigurationError(
                    f"No import for {dll}.{hint}.  Try by ordinal?"
                )
            imp = self._imports_by_name[(dll, hint)]
        elif isinstance(hint, int):
            # Look up by ordinal
            if (dll, hint) not in self._imports_by_ordinal:
                raise ConfigurationError(f"No import for {dll}.#{hint}.  Try by name?")
            imp = self._imports_by_ordinal[(dll, hint)]
        elif isinstance(hint, PEImport):
            imp = hint
        else:
            raise TypeError(f"Unexpected hint {hint} of type {type(hint)}")

        # Save the new value to the import
        # This marks it as initialized
        imp.value = value

        # Convert value to bytes
        if self.platform.byteorder == Byteorder.LITTLE:
            value_bytes = value.to_bytes(self.platdef.address_size, "little")
        elif self.platform.byteorder == Byteorder.BIG:
            value_bytes = value.to_bytes(self.platdef.address_size, "big")
        else:
            raise ConfigurationError(
                f"Can't encode int for byteorder {self.platform.byteorder}"
            )

        # Rewrite IAT slot
        # NOTE: PE files can have overlapping segments; check all of them.
        for off, seg in self.items():
            start = off + self.address
            stop = start + seg.get_size()
            if imp.iat_address >= start and imp.iat_address < stop:
                start = imp.iat_address - start
                end = start + len(value_bytes)
                contents = typing.cast(bytes, seg.get_content())
                contents = contents[0:start] + value_bytes + contents[end:]
                seg.set_content(contents)

    def link_pe(self, dll: "PEExecutable"):
        for imp in self._imports:
            if imp.value is not None:
                continue

            if imp.name is not None and (imp.dll, imp.name) in dll._exports_by_name:
                exp = dll._exports_by_name[(imp.dll, imp.name)]
            elif (
                imp.ordinal is not None
                and (imp.dll, imp.ordinal) in dll._exports_by_ordinal
            ):
                exp = dll._exports_by_ordinal[(imp.dll, imp.ordinal)]
            else:
                continue

            self.update_import(imp.dll, imp, exp.value)



import ctypes
import logging

import claripy
from angr.storage.memory_mixins.memory_mixin import MemoryMixin

from .... import hinting
from ....emulators.angr import PathTerminationSignal
from ....emulators.angr.utils import reg_name_from_offset
from ....extern.ctypes import TypedPointer
from .base import BaseMemoryMixin
from .visitor import EvalVisitor

log = logging.getLogger(__name__)
visitor = EvalVisitor()


class ModelMemoryMixin(BaseMemoryMixin):
    """
    Mixin for detecting model violations in memory ops.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._setup_tui()

    @MemoryMixin.memo
    def copy(self, memo):
        o = super().copy(memo)
        o._setup_tui()
        return o

    def _handle_typed_value(self, typedef, default, pretty):
        """
        Process user decision for a typed variable without a value

        Pointer types allow the following options:

        - Allocate a new region on the heap, and point to it.
        - Ase a reference to a new heap region

        All types allow the following options:

        - Use a placeholder symbol to track this value
        - Use a user-provided integer as a concrete value (not recommended for multi-KB structs)
        - Use zero/NULL as a concrete value
        - Use whatever default angr suggests
        - Print details of the current situation
        - Abort this execution path
        - Abort all execution
        """

        options = {
            "placeholder": self.typed_value_placeholder,
            "const": self.typed_value_const,
            "null": self.typed_value_null,
            "ignore": self.typed_value_ignore,
            "stop": self.typed_value_stop,
            "quit": self.typed_value_quit,
        }

        if issubclass(typedef, TypedPointer):
            option = "alloc"
            options.update(
                {
                    "alloc": self.typed_value_alloc,
                    "reuse": self.typed_value_reuse,
                }
            )
        else:
            option = "placeholder"
        return options[option](typedef=typedef, default=default, pretty=pretty)

    def _handle_typed_symbol(self, typedef, symbol, supercall, strategies, condition):
        # Symbols need special handling;
        options = {
            "default": self.typed_symbol_default,
            "const": self.typed_value_const,
            "null": self.typed_value_null,
            "stop": self.typed_value_stop,
            "quit": self.typed_value_quit,
        }

        if issubclass(typedef, TypedPointer):
            option = "alloc"
            options.update(
                {
                    "alloc": self.typed_value_alloc,
                    "reuse": self.typed_value_reuse,
                }
            )
        else:
            option = "default"
        return options[option](
            typedef=typedef,
            symbol=symbol,
            supercall=supercall,
            strategies=strategies,
            condition=condition,
        )

    def _handle_untyped_register(self, reg_name, default):
        options = {
            "ignore": self.untyped_value_ignore,
            "bind": self.untyped_addr_bind,
            "stop": self.untyped_value_stop,
            "quit": self.untyped_value_quit,
        }
        option = "ignore"
        return options[option](reg_name=reg_name, default=default)

    def _handle_untyped_address(self, addr, default):
        options = {
            "ignore": self.untyped_value_ignore,
            "bind": self.untyped_addr_bind,
            "stop": self.untyped_value_stop,
            "quit": self.untyped_value_quit,
        }
        option = "ignore"
        return options[option](addr=addr, default=default)

    def _handle_untyped_symbol(self, sym_name, default):
        options = {
            "ignore": self.untyped_value_ignore,
            "bind": self.untyped_sym_bind,
            "stop": self.untyped_value_stop,
            "quit": self.untyped_value_quit,
        }
        option = "ignore"
        return options[option](sym_name=sym_name, default=default)

    def _default_value(self, addr, size, **kwargs):
        environ = self.state.typedefs
        res = super()._default_value(addr, size, **kwargs)

        block = self.state.block()
        (insn,) = list(
            filter(
                lambda x: x.address == self.state._ip.concrete_value,
                block.disassembly.insns,
            )
        )
        if self.id == "reg":
            # Thanks to Pcode, this becomes much more complicated.
            # angr can't tell the difference between a read from a sub-register,
            # and read from a register truncated as part of the opcode's sleigh definition.
            # This makes it halucinate registers that don't actually exist.

            reg_name = reg_name_from_offset(self.state.arch, addr, size)
            reg_def = environ.get_register_binding(reg_name)
            if reg_def is not None:
                res = self._handle_typed_value(reg_def, res, reg_name)
                hint = hinting.TypedUnderSpecifiedRegisterHint(
                    message="Register has type, but no value",
                    typedef=str(reg_def),
                    register=reg_name,
                    instruction=self.state._ip.concrete_value,
                    value=str(res),
                )
                self.hinter.send(hint)
            else:
                res = self._handle_untyped_register(reg_name, res)
                hint = hinting.UntypedUnderSpecifiedRegisterHint(
                    message="Register has no type or value",
                    register=reg_name,
                    instruction=self.state._ip.concrete_value,
                    value=str(res),
                )
                self.hinter.send(hint)
            if isinstance(res, int):
                res = self.state.solver.BVV(res, size * 8)
        else:
            addr_def = environ.get_address_binding(addr)
            if addr_def is not None:
                res = self._handle_typed_value(addr_def, res, addr)
                hint = hinting.TypedUnderSpecifiedMemoryHint(
                    message="Memory has type, but no value",
                    typedef=str(addr_def),
                    address=addr,
                    size=size,
                    instruction=self.state._ip.concrete_value,
                    value=str(res),
                )
                self.hinter.send(hint)
            else:
                hint = hinting.UntypedUnderSpecifiedMemoryHint(
                    message="Memory has no type or value",
                    address=addr,
                    size=size,
                    instruction=self.state._ip.concrete_value,
                    value=str(res),
                )
                self.hinter.send(hint)
                self._handle_untyped_address(addr, res)

        return res

    def _concretize_bindings(self, supercall, addr, strategies, condition):
        """
        Concretization strategy for interrogating model

        This is the same code for both reads and writes.
        """
        environ = self.state.typedefs
        block = self.state.block()
        (insn,) = list(
            filter(
                lambda x: x.address == self.state._ip.concrete_value,
                block.disassembly.insns,
            )
        )

        log.debug(f"\tAddr:       {addr}")
        bindings = dict()
        complete = True
        for v in filter(lambda x: x.op == "BVS", addr.leaf_asts()):
            value = environ.get_symbol(v.args[0])
            if value is None:
                binding = environ.get_symbol_binding(v.args[0])
                if binding is None:
                    log.warn(
                        f"Symbol {v} (part of address expression {addr}) has no type"
                    )
                    value = self._handle_untyped_symbol(v, None)
                    if value is None:
                        complete = False
                        continue
                    hint = hinting.TypedUnderSpecifiedAddressHint(
                        message="Symbol has no type",
                        symbol=v.args[0],
                        addr=str(addr),
                        instruction=self.state._ip.concrete_value,
                        value=str(value),
                    )
                else:
                    log.warn(
                        f"Symbol {v} (part of address expression {addr}) has type {binding}, but no value"
                    )
                    while True:
                        value = self._handle_typed_symbol(
                            binding, v, supercall, strategies, condition
                        )
                        if value is not None and not self.state.solver.satisfiable(
                            extra_constraints=[v == value]
                        ):
                            log.warn(f"Selection {value:x} is not valid")
                            self._untyped_value_stop()
                        break
                    hint = hinting.TypedUnderSpecifiedAddressHint(
                        message="Symbol has type, but no value",
                        typedef=str(binding),
                        symbol=v.args[0],
                        addr=str(addr),
                        instruction=self.state._ip.concrete_value,
                        value=str(value),
                    )
                    self.hinter.send(hint)
                environ.set_symbol(v.args[0], value)
                if isinstance(value, int):
                    pretty_value = hex(value)
                else:
                    log.error(f"Bad value for {v.args[0]}: {value} ({type(value)})")
                    raise NotImplementedError()
                    quit(1)
                log.debug(f"\tNew Value for {v.args[0]}: {pretty_value}")
            else:
                log.debug(f"\tExisting Value for {v.args[0]}: {value:x}")
            bindings[v.args[0]] = claripy.BVV(value, 64)
        if complete:
            # We have bindings for all symbols.
            # These must be _concrete_ bindings.
            value = visitor.visit(addr, bindings=bindings)
            if not isinstance(value, int):
                value = value.concrete_value
            res = [value]
        else:
            log.warn(
                f"Address expression {addr} not completely concretized; falling back"
            )
            res = supercall(addr, strategies=strategies, condition=condition)
        log.debug("\tResults:")
        for v in res:
            if isinstance(v, int):
                log.debug(f"\t\t{v:x}")
            else:
                log.debug(f"\t\t{v}")
        return res

    def concretize_read_addr(self, addr, strategies=None, condition=None):
        log.debug("Concretizing on read:")
        return self._concretize_bindings(
            super().concretize_read_addr, addr, strategies, condition
        )

    def concretize_write_addr(self, addr, strategies=None, condition=None):
        log.debug("Concretizing on write:")
        return self._concretize_bindings(
            super().concretize_write_addr, addr, strategies, condition
        )

    # TUI handler functions.  Fear the boilerplate.

    def typed_value_alloc(self, typedef=None, **kwargs):
        # User wants to allocate a new instance
        environ = self.state.typedefs
        res = environ.allocate(typedef.type)
        log.warn(f"Allocated {ctypes.sizeof(typedef.type)} bytes at {res:x}")
        return res

    def typed_value_reuse(self, typedef=None, **kwargs):
        raise NotImplementedError("Reuse requires human input.")

    def typed_value_placeholder(self, typedef=None, **kwargs):
        # User wants to assign a placeholder symbol, not a real value.
        if hasattr(typedef, "__fields__"):
            log.info(f"{typedef} is struct-like")
            raise NotImplementedError(f"Not handling {typedef}")
        else:
            res = claripy.BVS(f"{typedef.__name__}", ctypes.sizeof(typedef) * 8)
            self.state.typedefs.bind_symbol(res, typedef)

        log.warn(f"Assigned placeholder {res}")
        return res

    def typed_value_const(self, **kwargs):
        # User wants to provide their own value
        res = input("( hex ) > ")
        res = int(res, 16)
        log.warn(f"Using user-provided value {res:x}")
        return res

    def typed_value_null(self, **kwargs):
        log.warn("Using NULL")
        return 0

    def typed_value_ignore(self, default=None, **kwargs):
        if default is None:
            log.warn("Will use default when determined.")
        elif isinstance(default, int):
            log.warn(f"Using default {default:x}")
        else:
            log.warn(f"Using default {default}")
        return default

    def typed_symbol_default(self, symbol, supercall, strategies, condition, **kwargs):
        res = supercall(symbol, strategies=strategies, condition=condition)[0]
        log.warn(f"Concretizing {symbol} to {res:x}")
        return res

    def typed_value_stop(self, **kwargs):
        log.warn("Stopping execution path")
        raise PathTerminationSignal()

    def typed_value_quit(self, **kwargs):
        log.warn("Aborting execution")
        quit()

    def untyped_value_ignore(self, default=None, **kwargs):
        if default is None:
            log.warn("Will use default when generated")
        elif isinstance(default, int):
            log.warn(f"Using default value {default:x}")
        else:
            log.warn(f"Using default value {default}")
        return default

    def untyped_reg_bind(**kwargs):
        raise NotImplementedError("Type binding not implemented for registers")

    def untyped_addr_bind(**kwargs):
        raise NotImplementedError("Type binding not implemented for addresses")

    def untyped_sym_bind(**kwargs):
        raise NotImplementedError("Type binding not implemented for symbols")

    def untyped_value_stop(self, **kwargs):
        self.warn("Stopping self execution path")
        raise PathTerminationSignal()

    def untyped_value_quit(self, **kwargs):
        log.warn("Aborting execution")
        quit()
